/*
   
 Numerical Recipes based FFT field filter
 
 filter_nr.cpp
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
 */

#include <iostream>
#include <stdlib.h>
#include <stdio.h>

// For Windows, we need to set the following to access M_PI
#if defined WIN32
#define _USE_MATH_DEFINES
#endif 

#include <math.h>

#include "sunmat.h"
#include "solve.h"

void filterFieldElement(double *FT3darray) {
	
	extern void FT3d(double* pArr, int nx, int ny,int nz, int sign);

	int fft_error = 0;
	//if ( NUMT != NUM ) fft_error = 1;
	if ( ! powerOfTwo(NUMT) ) fft_error = 2;
	
  if (fft_error > 0)
	{
    printError(fft_error);
    return;
 	}
	
  // seems all is ok, so take fft
  FT3d(FT3darray,NUMT,NUMT,NUM,1);
  
  // apply filter
  for (int sx=0;sx<NUMx;sx++) 
	  for (int sy=0;sy<NUMT;sy++)
      for (int sz=0; sz<NUM;sz++) {
      	float filter = filterFunction((float) sx/NUMT,(float) sy/NUMT,(float) sz/NUMT);
      	// filter real and imaginary parts
	      FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)] *= filter;
	      FT3darray[2*(sx*NUMT*NUM+sy*NUM+sz)+1] *= filter;  
	}

  // Inverse Fourier Transform
  FT3d(FT3darray,NUMT,NUMT,NUM,-1);
      	
  return;
}

void printError(int error) {
		cout << endl;
		cout << "   ****************************************************" << endl;
		cout << "   *                                                  *" << endl;
		cout << "   *  WARNING: FFT from Numerical Recipies does not   *" << endl;
		if (error == 1)
		cout << "   *  work with NUMT != NUM.                          *" << endl;
		if (error == 2)
		cout << "   *  work with NUMT, NUM not a strict power of two.  *" << endl;		
		cout << "   *                                                  *" << endl;
		cout << "   *  Can not apply field filters.                    *" << endl;
		cout << "   *  Either change NUMT and/or NUM or                *" << endl;
		cout << "   *  try making solve with FFT=FFTW'                 *" << endl;
		cout << "   *                                                  *" << endl;
		cout << "   ****************************************************" << endl;
		cout << endl;
}

