/*
 
 filter.h
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
*/

#ifndef __filter_h__
#define __filter_h__

#include "sunmat.h"

///////////////////////////////////////////////////////////
// Choose which group to use by default:                 //
#define FFT_NR  // Numerical recipies
// #define FFT_FFTW // FFTW
// Possible choices are: FFT_NR, FFT_FFTW
// !! ONLY DEFINE _ONE_ OF THESE !!                      //
///////////////////////////////////////////////////////////

// This choice can be overwritten by a global compiler switch:
#if defined useFFT_NR
#define FFT_NR
#undef FFT_FFTW
#elif defined useFFTW
#undef FFT_NR
#define FFT_FFTW
#endif

#if defined FFT_NR
#define FFT_string "FFT-NR"
#elif defined FFT_FFTW
#define FFT_string "FFTW"
#else
#error Please define here which source code to use.
#endif

//////////////////////////////////////////////////
// Constants
const bool MSG_FILTER = false; // Print Debug messages

//////////////////////////////////////////////////
// Common Functions
void filterFields();
void filterField(SUNmatrix*** field);
float filterFunction(float px, float py, float pz);

//////////////////////////////////////////////////
// Implemenation Specific Functions
void filterFieldElement(double *FT3darray);

//////////////////////////////////////////////////
// Utility Functions
bool powerOfTwo(int n);
void printError(int error);


#endif /* __filter_h__ */
