/*
 
 outputangle.h
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
*/

#ifndef __outputangle_h__
#define __outputangle_h__

#include "sunmat.h"

///////////////////////////////////////////////////////////
// Choose which group to use by default:                 //
#define FFT_NR  // Numerical recipies
// #define FFT_FFTW // FFTW
// Possible choices are: FFT_NR, FFT_FFTW
// !! ONLY DEFINE _ONE_ OF THESE !!                      //
///////////////////////////////////////////////////////////

// This choice can be overwritten by a global compiler switch:
#if defined useFFT_NR
#define FFT_NR
#undef FFT_FFTW
#elif defined useFFTW
#undef FFT_NR
#define FFT_FFTW
#endif

#if defined FFT_NR
#define FFT_string "FFT-NR"
#elif defined FFT_FFTW
#define FFT_string "FFTW"
#else
#error Please define here which source code to use.
#endif

//////////////////////////////////////////////////
// Constants
const bool MSG_FFT = false; // Print Debug messages
const bool ANGLE_MEASUREMENT_DNDPZ_MEAN = true; // Mean in dn dpz-measurement

//////////////////////////////////////////////////
// FFT-specific Functions
void fft_init();
void fft_end();

void gaugefix3d(int maxiter);
void measureangleforbins(double time, int usebins);
void measureangle(double time);

void outputSpectrum(SUNmatrix*** field, char* fieldName, double time);
void outputSpectra(double time);
double latticeK(int sx, int sy, int sz);

//////////////////////////////////////////////////
// Common Functions
extern SUNmatrix_BASIC_TYPE *loadsavebuffer;

void angledumpsave(double time);
int angledumpload(double time);
void measureanglefromdump();

#endif /* __ouputroutines_h__ */
