/*
 
 solve.h
 
 Copyright (C) 2008 Michael Strickland, Paul Romatschke, 
 Andreas Ipp, and Anton Rebhan
 
 Academic General Public License (AGPL)
 See detailed text in license directory
 
*/

#ifndef __solve_h__
#define __solve_h__

#include <sys/time.h>

#include "sunmat.h"

////////////////////////////////////////////////////
// Redo Angle measurement from dump:
//   This will not run the simulation, but
//   take dump files (especially created for angle
//   measurement) and process the angle measurement
//   with different bin settings.
//#define REDO_ANGLE_FROM_DUMP

////////////////////////////////////////////////////
// If ACCURATESUMMATION = true, we will minimize error in large summations
// (otherwise errors can be at percent level or higher!)
// [ ] not implemented in all functions yet.
const bool ACCURATESUMMATION = false;

#ifdef WIN32
// Windows only: Avoid deprecation warning for sprintf
#pragma warning(disable : 4996)
#endif

// external vars defined in solve.cpp
extern int NUM, NUMT, MIMICK1D, STEPS, UPDATE, SNAPUPDATE, ENERGYUPDATE;
extern int ANGLEUPDATE, ANGLEBINS1,  ANGLEBINS2,  ANGLEBINS3;
extern int CHECKPOINTUPDATE, TIMEOUT, STATISTICSUPDATE;
extern int USEPOLYHEDRA, NW, NPHI, NZ;
extern int RANDOMSEED, RANDOMSEEDUSED;
extern int PERTURB,MIMICKSUN,COUPLE_CURRENTS,FILTER_FUNC,INITIAL_CONDITION_TYPE;
// extern int NTRAJ, MIMICK2D, RANDOM, RANGEN, MIDPOINT; // not implemented yet?
extern double A, EPS, XI, MU2, SIG, PERTURBSIG, FILTER_WIDTH, MAXEVIOLATION; // Unneeded, historical?:, TSIG, TINIT;

// is set below to NUMx = NUMT, but smaller in parallel application.
extern int NUMx; 

// these hold the current values; note w[v] = a[v] wo[v] + b[v] wz[v]  
extern SUNmatrix ***pix, ***piy, ***ez, ****w;
extern SUNLieMatrix ***u, ***ux, ***uy;

// these hold the updated values
extern SUNmatrix ***Pix, ***Piy, ***Ez, ****W;
extern SUNLieMatrix ***U, ***Ux, ***Uy;

// these hold the old values of W used for making it's 
// time derivative symmetric in t
extern SUNmatrix ****ow;

// j holds current at t; oj holds current at t-eps
extern SUNmatrix ****j, ****oj;

// vertex array - x=0,y=1,z=2 - and coefficient arrays - a and b
extern double **v, *a, *b;

// InitialEnergy of the system 
extern double InitialEnergy;

// Holds the percentage difference from the initial energy
extern double energyDiff;

// stores the running integrated value of the HL contribution to the energy
extern double HLenergy;

// stores timing information (time struct as used by gettimeofday())
extern timeval starttime;  // time at program start
extern timeval outputtime; // time of last outputmeasurement()
extern long outputcounter; // counts number of outputmeasurements

void initializeVelocitiesAndCoefficients();
void enforceBCs();
void enforceBCsOnlyforPU();
void setInitialConditionsForSlice(int sx);
void perturbInitialConditionsForSlice(int sx);
void loadjs();
void savejs();
void doIncPart1(double eps);
void doIncPart2(double eps);
void updateHLenergy(double eps);
double ezEnergyDensity();
double etEnergyDensity();
double bzEnergyDensity();
double btEnergyDensity();
SUNLieMatrix dPlaqXY(int sx, int sy, int sz);
SUNLieMatrix dPlaqXZ(int sx, int sy, int sz);
SUNLieMatrix dPlaqYZ(int sx, int sy, int sz);
double initialED();
double energyDensity();
SUNmatrix checkGaussLawPart1();
double checkGaussLawPart2(SUNmatrix temp);
double checkGaussLaw();
double newCjbar();
double Cjxzbar();
double Cjyzbar();
void printDivider();

int solve();

#endif /* __solve_h__ */
