#ifndef __solvempi_h__
#define __solvempi_h__

/////////////////////////////////////////////////////////////////////
//
// Choose which output messages to display:
//
/////////////////////////////////////////////////////////////////////
//
// (set all to "true" if you want to see how the program works!)
//
const bool MSG_MPIINIT = true; // MPI initialization related messages

const bool MSG_ALL = false; // Set true to set all messages below to true.

const bool MSG_OUTPUT = false || MSG_ALL; // Output Measurement related messages
const bool MSG_SENDBUFFER = false || MSG_ALL; // Send Buffer related messages
const bool MSG_GARBAGE = false || MSG_ALL; // Garbage collection related messages
const bool MSG_INFO = false || MSG_ALL; // Information related messages
const bool MSG_SENDDATA = false || MSG_ALL; // Sending data related messages
const bool MSG_STATEDUMP = false || MSG_ALL; // Provide state dumps
const bool MSG_MESSAGE = false || MSG_ALL; // Message transmission related messages
const bool MSG_PROGRAM = false || MSG_ALL; // Program execution related messages
const bool MSG_DYNAMIC = false || MSG_ALL; // Dynamic work load related messages
const bool MSG_IDLE = false || MSG_ALL; // whether to indicate idle states.
const bool MSG_DISK = false || MSG_ALL; // disk related messages.
// (Default: Only MSG_MPIINIT = true)
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
//
// Set additinal parameters:
//
/////////////////////////////////////////////////////////////////////
// Use dynamic load balancing:
// * false: each computer (node) only works on those slices 
//          it received at the beginning of the simulation.
// * true:  borders between the slices are moved at each time-step, 
//          and the work load is dynamically balanced among 
//          the computers (nodes).
const bool OPTION_DYNAMICLOADBALANCING = false;
/////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// Choose Memory optimizations:
// Drop Memory requirements for OW and OJ fields:
// All can be simulated with pointers.
const bool OPTION_MEMORY_DROP_OW_OJ = true;
/////////////////////////////////////////////////////////////////////

const int DEFAULT_SLEEP_MS = 10; // Default sleep-time in milli-seconds when idle.

/////////////////////////////////////////////////////////////////////


extern int MPIsize;


/////////////////////////////////////////////////////////////////////
//
// Definition of tags used to identify MPI messages
//
/////////////////////////////////////////////////////////////////////

const int TAG_DATA = 10;  // All data for one slice of the simulation.
const int TAG_DIE = 11; // Finish jobs.

// Initial distribution of border cells:
const int TAG_ALLSLICESSENT = 12; // Manager sent all available slices
const int TAG_ALLSLICESRECEIVED = 13; // Acknowledgement from workers.
const int TAG_STARTSIMULATION = 14; // Finish by shakehand.
const int TAG_REACHEDSIMULATIONEND = 15; // Reached end of simulation.
const int TAG_SETTIME = 16; // Set time when starting from data dump

const int TAG_WANTTOSEND = 20; // Indicate that we have data to send
const int TAG_PLEASESEND = 21; // Buffer prepared, please send
const int TAG_DONOTSEND = 22; // I'm talking to someone else, please don't send
const int TAG_DATABLOCK = 23; // Contains asynchronous block of data

const int TAG_RESETFIRSTSLICEPOINTER = 30; // Special treatment for last node
const int TAG_INCREMENTLASTSLICEPOINTER = 31; // Special treatment for last node
const int TAG_FINISHEDSLICE = 32; // Worker notifies Manager about finished work
const int TAG_TIMEOUT = 33; // Manager sends timeout signal
const int TAG_RESETSLICEPOINTERS = 34; // Special treatment for data dump

// BM (BorderMove) related tags for dynamic work loading:
const int TAG_BM_PLEASESENDTWO_TOLEFT = 40; // Border move: send me two, I win.
const int TAG_BM_PLEASESENDTWO_TORIGHT = 41; // Border move: send me two, I win.
const int TAG_BM_SENDINGTWO_TOLEFT = 42; // Ok, I send 2 slices.
const int TAG_BM_SENDINGTWO_TORIGHT = 43; // Ok, I send 2 slices.
const int TAG_BM_SENDINGONE_TOLEFT = 44; // Not ok
const int TAG_BM_SENDINGONE_TORIGHT = 45; // Not ok

const int TAG_OUTPUTVARIABLES = 50; // Send set of output variables
const int TAG_STATISTICSVARIABLES = 51; // Send set of statistics variables

/////////////////////////////////////////////////////////////////////
//
//     States and Constants
//
/////////////////////////////////////////////////////////////////////

// Send states
const int STATE_NOTSENDING = 100; // Not sending
const int STATE_WAITINGTOSEND = 101; // Waiting to send
const int STATE_SENDING = 102;

// Receive states
const int STATE_NOTRECEIVING = 200;
const int STATE_RECEIVING = 201;

// Dynamic load balancing states (BM..Border Move)
const int STATE_BM_NOTWAITING = 400;
const int STATE_BM_WAITINGFORRESPONSE = 401;

// If we refer to all parts at the same time.
// Parts are marked bitwise (because we might get them in 
// arbitrary order, and we should still know which of
// the parts we have or not)
const int PART1 = 1;
const int PART2 = 2;
const int ALL_PARTS = PART1 | PART2;

// Which parts in a Simulation slice to allocate:
const int FIELDS_PU = 1; // contains fields pix, piy, ez, ux, uy, u
const int FIELDS_WJ = 2; // contains fields w, j
const int ALL_FIELDS = FIELDS_PU | FIELDS_WJ;
const int NO_FIELDS = 0;

// DataDump state
const int DATADUMP_NOTHING = 0;
const int DATADUMP_FIRST = 1;
const int DATADUMP_MIDDLE = 2;
const int DATADUMP_LAST = 3;

#endif /* __solvempi_h__ */
