/*
   Header file to solve.cpp

   AI, March 11, 2006
*/
#ifndef __sunmat_h__
#define __sunmat_h__

#include <iostream>
#include <iomanip>
#include <fstream>
#include <math.h>

///////////////////////////////////////////////////////////
// Choose which group to use by default:                 //
//#define SU2
#define SU2double
//#define SU3
//#define SU3double
//#define SU3tvmet
// Possible choices are: SU2, SU3, SU3double, SU3tvmet
// !! ONLY DEFINE _ONE_ OF THESE !!                      //
///////////////////////////////////////////////////////////

// This choice can be overwritten by a global compiler switch:
#if defined useSU2
#define SU2
#undef SU3
#undef SU3double
#undef SU3tvmet
#elif defined useSU2double
#undef SU2
#define SU2double
#undef SU3
#undef SU3double
#undef SU3tvmet
#elif defined useSU3
#undef SU2
#define SU3
#undef SU3double
#undef SU3tvmet
#elif defined useSU3double
#undef SU2
#undef SU3
#define SU3double
#undef SU3tvmet
#elif defined useSU3tvmet
#undef SU2
#undef SU3
#undef SU3double
#define SU3tvmet
#endif

/* Note:
*
*  If you add a group, you have to manually update:
*
*	in solve.cpp:
*		void setInitialConditionsForSlice(int sx);  [ ] factor correct????????????
*
*	in solvempi.cpp, class SimulationSlice:
*		void packSUNmatrixarray(SUNmatrix *p, int max);
*		void unpackSUNmatrixarray(SUNmatrix *p, int max);
*		void packSUNLieMatrixarray(SUNLieMatrix *p, int max);
*		void unpackSUNLieMatrixarray(SUNLieMatrix *p, int max);
*
*/

#if defined SU2

#include "su2mat.h"
typedef float                       SUNmatrix_BASIC_TYPE;
typedef double                      SUNLieMatrix_BASIC_TYPE;
typedef SU2matrix<float>            SUNmatrix;
typedef SU2LieMatrix<float, double> SUNLieMatrix;
#define SUNmatrix_string "SU(2)"
//const char* SUNmatrix_string = "SU(2)";
const int SUNmatrix_NUMBER_OF_FLOATS  = 3; // 3 float in SU2matrix
const int SUNmatrix_NUMBER_OF_DOUBLES = 0; // 0 double in SU2matrix
const int SUNLieMatrix_NUMBER_OF_FLOATS = 0; // 0 doubles in SU2LieMatrix
const int SUNLieMatrix_NUMBER_OF_DOUBLES = 4; // 4 doubles in SU2LieMatrix
inline void SUNLieMatrix_RESET(SUNLieMatrix m) {m.a = 0; m.b = 0; m.c = 0; m.d = 0;}
inline void SUNmatrix_RESET(SUNmatrix m) {m.a = 0; m.b = 0; m.c = 0;}
inline void SUNLieMatrix_OUT(fstream &s, SUNLieMatrix m)
{
	s << m.a << "\t" << m.b << "\t" << m.c << "\t" << m.d;
};
inline void SUNmatrix_OUT(fstream &s, SUNmatrix m)
{
	s << m.a << "\t" << m.b << "\t" << m.c;
};

#elif defined SU2double

#include "su2mat.h"
typedef float                        SUNmatrix_BASIC_TYPE;
typedef double                       SUNLieMatrix_BASIC_TYPE;
typedef SU2matrix<double>            SUNmatrix;
typedef SU2LieMatrix<double, double> SUNLieMatrix;
#define SUNmatrix_string "SU(2)"
//const char* SUNmatrix_string = "SU(2)double";
const int SUNmatrix_NUMBER_OF_FLOATS  = 0; // 0 float in SU2matrix
const int SUNmatrix_NUMBER_OF_DOUBLES = 3; // 3 double in SU2matrix
const int SUNLieMatrix_NUMBER_OF_FLOATS = 0; // 0 doubles in SU2LieMatrix
const int SUNLieMatrix_NUMBER_OF_DOUBLES = 4; // 4 doubles in SU2LieMatrix
inline void SUNLieMatrix_RESET(SUNLieMatrix m) {m.a = 0; m.b = 0; m.c = 0; m.d = 0;}
inline void SUNmatrix_RESET(SUNmatrix m) {m.a = 0; m.b = 0; m.c = 0;}
inline void SUNLieMatrix_OUT(fstream &s, SUNLieMatrix m)
{
	s << m.a << "\t" << m.b << "\t" << m.c << "\t" << m.d;
};
inline void SUNmatrix_OUT(fstream &s, SUNmatrix m)
{
	s << m.a << "\t" << m.b << "\t" << m.c;
};

#elif defined SU3

#include "su3mat.h"
typedef float                                 SUNmatrix_BASIC_TYPE;
typedef double                                SUNLieMatrix_BASIC_TYPE;
typedef SU3matrix<float>                      SUNmatrix;
typedef SU3LieMatrix<float, complex<double> > SUNLieMatrix;
#define SUNmatrix_string "SU(3)"
const int SUNmatrix_NUMBER_OF_FLOATS  = 8; // 8 float in SU3matrix
const int SUNmatrix_NUMBER_OF_DOUBLES = 0; // 0 double in SU3matrix
const int SUNLieMatrix_NUMBER_OF_FLOATS = 0; // 0 doubles in SU3LieMatrix
const int SUNLieMatrix_NUMBER_OF_DOUBLES = 18; // 18 doubles in SU3LieMatrix
inline void SUNLieMatrix_RESET(SUNLieMatrix m) {m.reset();}
inline void SUNmatrix_RESET(SUNmatrix m) {m.reset();}
inline void SUNLieMatrix_OUT(fstream &s, SUNLieMatrix m)
{
	s << real(m.m11) << " + I " << imag(m.m11) << "\t";
	s << real(m.m12) << " + I " << imag(m.m12) << "\t";
	s << real(m.m13) << " + I " << imag(m.m13) << "\t";
	s << real(m.m21) << " + I " << imag(m.m21) << "\t";
	s << real(m.m22) << " + I " << imag(m.m22) << "\t";
	s << real(m.m23) << " + I " << imag(m.m23) << "\t";
	s << real(m.m31) << " + I " << imag(m.m31) << "\t";
	s << real(m.m32) << " + I " << imag(m.m32) << "\t";
	s << real(m.m33) << " + I " << imag(m.m33);
};
inline void SUNmatrix_OUT(fstream &s, SUNmatrix m)
{
  /*
        // Output as LieMatrix (compatible to tvmet)
	//SUNLieMatrix lm(m);
	SUNLieMatrix lm;
	lm.algebra_to_liegroup(m);
	SUNLieMatrix_OUT(s, lm);
  */
	// Direct output of components:
	s << "\t||\t";
	s << m.a1 << "\t";
	s << m.a2 << "\t";
	s << m.a3 << "\t";
	s << m.a4 << "\t";
	s << m.a5 << "\t";
	s << m.a6 << "\t";
	s << m.a7 << "\t";
	s << m.a8;

};

#elif defined SU3double

#include "su3mat.h"
typedef double                                 SUNmatrix_BASIC_TYPE;
typedef double                                 SUNLieMatrix_BASIC_TYPE;
typedef SU3matrix<double>                      SUNmatrix;
typedef SU3LieMatrix<double, complex<double> > SUNLieMatrix;
#define SUNmatrix_string "SU(3)double"
const int SUNmatrix_NUMBER_OF_FLOATS  = 0; // 0 float in SU3matrix
const int SUNmatrix_NUMBER_OF_DOUBLES = 8; // 8 double in SU3matrix
const int SUNLieMatrix_NUMBER_OF_FLOATS = 0; // 0 doubles in SU3LieMatrix
const int SUNLieMatrix_NUMBER_OF_DOUBLES = 18; // 18 doubles in SU3LieMatrix
inline void SUNLieMatrix_RESET(SUNLieMatrix m) {m.reset();}
inline void SUNmatrix_RESET(SUNmatrix m) {m.reset();}
inline void SUNLieMatrix_OUT(fstream &s, SUNLieMatrix m)
{
	s << real(m.m11) << " + I " << imag(m.m11) << "\t";
	s << real(m.m12) << " + I " << imag(m.m12) << "\t";
	s << real(m.m13) << " + I " << imag(m.m13) << "\t";
	s << real(m.m21) << " + I " << imag(m.m21) << "\t";
	s << real(m.m22) << " + I " << imag(m.m22) << "\t";
	s << real(m.m23) << " + I " << imag(m.m23) << "\t";
	s << real(m.m31) << " + I " << imag(m.m31) << "\t";
	s << real(m.m32) << " + I " << imag(m.m32) << "\t";
	s << real(m.m33) << " + I " << imag(m.m33);
};
inline void SUNmatrix_OUT(fstream &s, SUNmatrix m)
{
  /*
        // Output as LieMatrix (compatible to tvmet)
	SUNLieMatrix lm;
	lm.algebra_to_liegroup(m);
	SUNLieMatrix_OUT(s, lm);
  */
	// Direct output of components:
	s << "\t||\t";
	s << m.a1 << "\t";
	s << m.a2 << "\t";
	s << m.a3 << "\t";
	s << m.a4 << "\t";
	s << m.a5 << "\t";
	s << m.a6 << "\t";
	s << m.a7 << "\t";
	s << m.a8;

};

#elif defined SU3tvmet

#include "su3mat-tvmet.h"
typedef double    SUNmatrix_BASIC_TYPE;
typedef double    SUNLieMatrix_BASIC_TYPE;
typedef SU3matrix SUNmatrix;
typedef SU3matrix SUNLieMatrix;
#define SUNmatrix_string "SU(3)-tvmet"
const int SUNmatrix_NUMBER_OF_FLOATS  = 0; // 0 float in SU3matrix
const int SUNmatrix_NUMBER_OF_DOUBLES = 18; // 18 double in SU3matrix
const int SUNLieMatrix_NUMBER_OF_FLOATS = 0; // 0 doubles in SU3matrix
const int SUNLieMatrix_NUMBER_OF_DOUBLES = 18; // 18 doubles in SU3matrix
inline void SUNLieMatrix_RESET(SUNLieMatrix m) {m.mat = 0;}
inline void SUNmatrix_RESET(SUNmatrix m) {m.mat = 0;}
inline void SUNLieMatrix_OUT(fstream &s, SUNLieMatrix m)
{
	for(int i=0;i<3;i++)
	{
		for(int j=0;j<3;j++)
		{
			s << real(m.mat(i,j)) << " + I " << imag(m.mat(i,j));
			if ((i<2)||(j<2)) s << "\t";
		};
	};
};
inline void SUNmatrix_OUT(fstream &s, SUNmatrix m)
{
	SUNLieMatrix_OUT(s, m);
};

#else
#error Please specify a group SU(N)! // Throw compiler error
#endif

#endif // __sunmat_h__
