/*
  Automatically include the right outputangle code
  + common routines
  Andreas Ipp, July 19, 2006
*/

#include "outputangle.h"

////////////////////////////////////////////////////////////////////////
// Include FFT specific code:
#if defined FFT_NR
#include "outputangle_nr.cpp"
#elif defined FFT_FFTW
#include "outputangle_fftw.cpp"
#else
#error Please define here which source code to use.
#endif


////////////////////////////////////////////////////////////////////////
// Common routines to both codes:


////////////////////////////////////////////////////////////////////////
//
// Data dump save & load

SUNmatrix_BASIC_TYPE *loadsavebuffer;  // This is set at fft_init()

static void save_field(fstream *pout, SUNmatrix ***field)
{
	int SUNmax = field[1][1][1].maxelement();
	for (int SUNelement = 0; SUNelement < SUNmax; SUNelement++)
	{
		// Write field into buffer:
		int n = 0;
		for (int sx=1;sx<=NUMT;sx++)
		for (int sy=1;sy<=NUMT;sy++)
		for (int sz=1;sz<=NUM;sz++)
		{
			loadsavebuffer[n] = field[sx][sy][sz].element(SUNelement);
			n++;
		}
		assert( n == NUMT * NUMT * NUM );

		// write to file
		(*pout).write((char *)loadsavebuffer, sizeof(loadsavebuffer[1]) * n);
	};

}

static void load_field(ifstream *pin, SUNmatrix ***field)
{
	// cout << "Reset fields" << endl;

	// Reset the fields
	for (int sx=1;sx<=NUMT;sx++)
	for (int sy=1;sy<=NUMT;sy++)
	for (int sz=1;sz<=NUM;sz++)
	{
		field[sx][sy][sz].reset();
	};

	// cout << "Load fields from file" << endl;
	int SUNmax = field[1][1][1].maxelement();
	for (int SUNelement = 0; SUNelement < SUNmax; SUNelement++)
	{
		// load from file
		(*pin).read((char *)loadsavebuffer, sizeof(loadsavebuffer[1]) * NUMT * NUMT * NUM);

		// Load field from buffer:
		int n = 0;
		for (int sx=1;sx<=NUMT;sx++)
		for (int sy=1;sy<=NUMT;sy++)
		for (int sz=1;sz<=NUM;sz++)
		{
			field[sx][sy][sz].addelement(SUNelement, loadsavebuffer[n]);
			n++;
		}
		assert( n == NUMT * NUMT * NUM );
	};
}

// Dump data currently in the buffer to a file.
void angledumpsave(double time)
{
    fstream out;
    char filename[255];

	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	if (MSG_FFT)
		cout << "0: Dumping angle measurement data. time: " << time << endl;
	
	sprintf(filename,"%s/snapshot/fdump_%.2f.dmp",pdirname,time);

	// file name defined above
	out.open(filename, ios::out|ios::binary);

	save_field(&out, pix); 
	save_field(&out, piy); 
	save_field(&out, ez); 

	out.close();

};

// Load data dump from file.
// returns 0 if successful, otherwise 1.
int angledumpload(double time)
{
	int status;
    ifstream in;
    char filename[255];
	
	char *pdirname = "data";
	if (NUMx == 1) pdirname = "data_mpi";

	if (MSG_FFT)
		cout << "0: Loading angle measurement data. time: " << time << endl;

	sprintf(filename,"%s/snapshot/fdump_%.2f.dmp",pdirname,time);

	// file name defined above
	in.open(filename, ios::out|ios::binary);
	if (in.fail() == 1)
	{
		// File does not exist
		status = 1; // File does not exist
	}
	else
	{
		load_field(&in, pix); 
		load_field(&in, piy); 
		load_field(&in, ez); 
		status = 0; // Everything ok  ([ ] we didn't check for corrupt files...)
	};
	in.close();

	return status;
};

// This routine reads in data from (existing) angle dump files,
// performs the measurement for a specified number of bins,
// and saves the output data.
void measureanglefromdump()
{
	// Programming remark: 
	// the loop structure of this function follows Evolve()
	double eps = EPS;
	double time = 0;
	int filemiss = 0;  // Number of filemisses.

	cout 
		<< "  ++++++++++++++++++++++++++++++++++++++++++++++" << endl
		<< "  +                                            +" << endl
		<< "  +   Not doing calculation but just           +" << endl
		<< "  +   performing angle measurement.            +" << endl
		<< "  +                                            +" << endl
		<< "  ++++++++++++++++++++++++++++++++++++++++++++++" << endl 
		<< endl;

	for (int i=1;i<=STEPS;i++)
	{
		if ((i-1)%ANGLEUPDATE==0)
		{
			// Load angle dump:
			if (angledumpload(time) == 0)
			{
				cout << "Processing at time " << time << endl;
				// Successfully loaded
				// cout << "Measuring time " << time << endl;
				measureangleforbins(time, ANGLEBINS1);
				measureangleforbins(time, ANGLEBINS2);
				measureangleforbins(time, ANGLEBINS3);
			}
			else
			{
				// Could not load data
				filemiss++;
				cout << "Miss " << filemiss << ": dump for time " << time << " does not exist." << endl;
				if (filemiss >= 5) break; // for
			};
		}
		// increment time
		time = i*eps;
	};
	cout << "Angle measurements done." << endl;
};

