#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include "outputroutines.h"
#include "sunmat.h"
#include "solve.h"

// because this is included in solve.cpp you should 'touch solve.cpp' 
// after edits to this in order for make to know there has been a change

// => MPI implementation uses outputMeasurementsMPIworker() and outputMeasurementsMPImanager().
void outputMeasurements(float time) {

	double cjxybar = newCjbar();
	double cjxzbar = Cjxzbar();
	double cjyzbar = Cjyzbar();
	
	double etenergy = etEnergyDensity();
	double ezenergy = ezEnergyDensity();
	double bzenergy = bzEnergyDensity();
	double btenergy = btEnergyDensity();
	double hlenergy = HLenergy/NUM/NUMT/NUMT/A/A/A/A;

	double energy = 0;
	if (ACCURATESUMMATION)
	{
		double denergy = 0;
		denergy += hlenergy;
		denergy += ezenergy;
		denergy += etenergy;
		denergy += bzenergy;
		denergy += btenergy;
		energy = denergy;
	}
	else
	{
		energy += hlenergy;
		energy += ezenergy;
		energy += etenergy;
		energy += bzenergy;
		energy += btenergy;
	}

	//energyDiff = (energy-InitialEnergy)/InitialEnergy;
	//energyDiff=energy
  energyDiff = energy-InitialEnergy;
	double gausslaw = checkGaussLaw();

	// Update timing information
	timeval newtime;
	double outputseconds, averageseconds;
	outputcounter++;
	gettimeofday(&newtime, 0);
	outputseconds = ((double)(newtime.tv_sec - outputtime.tv_sec)) 
	  + ((double)(newtime.tv_usec - outputtime.tv_usec) * 0.000001);
	averageseconds = (((double)(newtime.tv_sec - starttime.tv_sec)) 
	  + ((double)(newtime.tv_usec - starttime.tv_usec) * 0.000001)) / outputcounter;
        outputtime = newtime;
		
	// output to screen

	cout.precision(5);
	int dwidth = 13;
	cout.width(dwidth); cout << time;
	cout.width(dwidth); cout << ezenergy;
	cout.width(dwidth); cout << etenergy;
	cout.width(dwidth); cout << bzenergy;
	cout.width(dwidth); cout << btenergy;
	cout.width(dwidth); cout << hlenergy;
	cout.width(dwidth); cout << gausslaw;
	cout.width(dwidth); cout << energyDiff;
	cout.width(dwidth); cout << outputseconds;
	cout.width(dwidth); cout << averageseconds;
	cout << endl;

	// output to files

	cbar_out << time;
	if (ACCURATESUMMATION)
	{
		cbar_out << "\t" << log10((double) cjxybar);
		cbar_out << "\t" << log10((double) cjxzbar);
		cbar_out << "\t" << log10((double) cjyzbar) << endl;
	}
	else
	{
		cbar_out << "\t" << log10(cjxybar);
		cbar_out << "\t" << log10(cjxzbar);
		cbar_out << "\t" << log10(cjyzbar) << endl;
	};

	gausslaw_out << time << "\t" << log10(gausslaw) << endl; 

	energy_out << time;
	energy_out << "\t" << fabs(energyDiff);
	energy_out << "\t" << ezenergy;
	energy_out << "\t" << etenergy;
	energy_out << "\t" << bzenergy;
	energy_out << "\t" << btenergy;
	energy_out << "\t" << fabs(hlenergy);
	energy_out << "\t" << energy*A*A*A*A/3;
	energy_out << endl;
}

void dumpEnergyDensities(float time) {

  fstream out;
  char fname[255];
  int N = SUNLieMatrix::N();
  	
	char *pdirname = "data";
  if (NUMx == 1) pdirname = "data_mpi";
  
	// dump un-PT'ed j's, x,y const;
	
	sprintf(fname,"%s/snapshot/energyDensity_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
  out.precision(20);
	for (int sx=1;sx<=NUMx;sx++)
	for (int sy=1;sy<=NUMT;sy++)
	for (int sz=1;sz<=NUM;sz++)
	{
		  out << tr(dPlaqXY(sx,sy,sz))/N;
		  out << "\t";
			out << tr(dPlaqXZ(sx,sy,sz))/N;
			out << "\t";
			out << tr(dPlaqYZ(sx,sy,sz))/N;
	    out << endl;
  }			
  out.close();
}

// MPI implementation note: 
// Due to memory constraints,
// ONLY data at x==NUMT/2, y and z arbitrary
// and data at y==NUMT/2, x and y arbitrary
// are provided in the MPI implementation.
// You MUST NOT probe at other points.
void snapshot(float time) {

  int h=NUM/2;
  int ht=NUMT/2;
  
  SUNLieMatrix dPlaqxy,dPlaqxz,dPlaqyz;

  fstream out;
  char fname[255];

  char *pdirname = "data";
  if (NUMx == 1) pdirname = "data_mpi";
  
	// dump un-PT'ed j's, x,y const;
	
	sprintf(fname,"%s/snapshot/j_z_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1-1;s<=NUM+1;s++) { 
		out << s << "\t";
//		out << j[ht][ht][s][0].a << "\t";
//		out << j[ht][ht][s][0].b << "\t";
//		out << j[ht][ht][s][0].c << "\t";
		SUNmatrix_OUT(out, j[ht][ht][s][0]);
		out << "\t";
		SUNmatrix_OUT(out, j[ht][ht][s][1]);
		out << "\t";
		SUNmatrix_OUT(out, j[ht][ht][s][2]);
		out << endl << endl;
	}
	out.close();

	// dump j's pt'd, x,y const;

	sprintf(fname,"%s/snapshot/jPT_z_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUM;s++) { 
		out << s << "\t";
		SUNmatrix jxPT = j[ht][ht][s][0];
		SUNmatrix jyPT = j[ht][ht][s][1];
		SUNmatrix jzPT = j[ht][ht][s][2];
		for (int i=s; i>=2; i--) {
		  // this is confusing due to a redefinition of left and right between
		  // 1+1 and 3+1 versions.  This is apparently transporting to the left
		  // but the function name is highly confusing and should be fixed in 
		  // order to avoid future confusion.
		  jxPT = transportRight(jxPT,u[ht][ht][i-1]);
		  jyPT = transportRight(jyPT,u[ht][ht][i-1]);
		  jzPT = transportRight(jzPT,u[ht][ht][i-1]);
		}
		SUNmatrix_OUT(out, jxPT);
		out << "\t";
		SUNmatrix_OUT(out, jyPT);
		out << "\t";
		SUNmatrix_OUT(out, jzPT);
		out << endl << endl;
	}
	out.close();

	// dump unpted j's, x,z const;
	
	sprintf(fname,"%s/snapshot/j_y_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1-1;s<=NUMT+1;s++) { 
		out << s << "\t";
		SUNmatrix_OUT(out, j[ht][s][h][0]);
		out << "\t";
		SUNmatrix_OUT(out, j[ht][s][h][1]);
		out << "\t";
		SUNmatrix_OUT(out, j[ht][s][h][2]);
		out << endl << endl;
	}
	out.close();

	// dump pt'd j's, x,z const;

	sprintf(fname,"%s/snapshot/jPT_y_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUMT;s++) { 
	  out << s << "\t";
	  SUNmatrix jxPT = j[ht][s][h][0];
	  SUNmatrix jyPT = j[ht][s][h][1];
	  SUNmatrix jzPT = j[ht][s][h][2];
	  for (int i=s; i>=2; i--) {
	    jxPT = transportRight(jxPT,uy[ht][i-1][h]);
	    jyPT = transportRight(jyPT,uy[ht][i-1][h]);
	    jzPT = transportRight(jzPT,uy[ht][i-1][h]);
	  }
	  SUNmatrix_OUT(out, jxPT);
	  out << "\t";
	  SUNmatrix_OUT(out, jyPT);
	  out << "\t";
	  SUNmatrix_OUT(out, jzPT);
	  out << endl << endl;
	}
	out.close();


	// dump unpted j's, y,z const;

	sprintf(fname,"%s/snapshot/j_x_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUMT;s++) { 
		out << s << setprecision(20) << "\t";
	    SUNmatrix_OUT(out, j[s][ht][h][0]);
	    out << "\t";
	    SUNmatrix_OUT(out, j[s][ht][h][1]);
	    out << "\t";
	    SUNmatrix_OUT(out, j[s][ht][h][2]);
		out << endl << endl;
	}
	out.close();
	// dump pt'd j's, y,z const;

	sprintf(fname,"%s/snapshot/jPT_x_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUMT;s++) { 
	  out << s << "\t";
	  SUNmatrix jxPT = j[s][ht][h][0];
	  SUNmatrix jyPT = j[s][ht][h][1];
	  SUNmatrix jzPT = j[s][ht][h][2];

	  for (int i=s; i>=2; i--) {
	    jxPT = transportRight(jxPT,ux[i-1][ht][h]);
	    jyPT = transportRight(jyPT,ux[i-1][ht][h]);
	    jzPT = transportRight(jzPT,ux[i-1][ht][h]);
	  }

	    SUNmatrix_OUT(out, jxPT);
	    out << "\t";
	    SUNmatrix_OUT(out, jyPT);
	    out << "\t";
	    SUNmatrix_OUT(out, jzPT);
		out << endl << endl;
	}
	out.close();

	// dump electric fields, pi's (ex and ey) and ez's, x,y const;

	sprintf(fname,"%s/snapshot/e_z_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUM;s++) { 
		out << s << "\t";
	    SUNmatrix_OUT(out, pix[ht][ht][s]);
	    out << "\t";
	    SUNmatrix_OUT(out, piy[ht][ht][s]);
	    out << "\t";
	    SUNmatrix_OUT(out, ez[ht][ht][s]);
		out << endl << endl;
	}
	out.close();	

	// dump electric fields, pi's (ex and ey) and ez's, x,z const;

	sprintf(fname,"%s/snapshot/e_y_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUMT;s++) { 
		out << s << "\t";
	    SUNmatrix_OUT(out, pix[ht][s][h]);
	    out << "\t";
	    SUNmatrix_OUT(out, piy[ht][s][h]);
	    out << "\t";
	    SUNmatrix_OUT(out, ez[ht][s][h]);
		out << endl << endl;
	}
	out.close();	

	// dump electric fields, pi's (ex and ey) and ez's, y,z const;
	
	sprintf(fname,"%s/snapshot/e_x_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);
	for (int s=1;s<=NUMT;s++) { 
		out << s << "\t";
	    SUNmatrix_OUT(out, pix[s][ht][h]);
	    out << "\t";
	    SUNmatrix_OUT(out, piy[s][ht][h]);
	    out << "\t";
	    SUNmatrix_OUT(out, ez[s][ht][h]);
		out << endl << endl;
	}
	out.close();	

	// dump B's, x,y const;

	sprintf(fname,"%s/snapshot/b_z_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);	
	for (int s=1;s<=NUM;s++) {
	  dPlaqxy=ux[ht][ht][s]*uy[ht+1][ht][s]*Hermitian(ux[ht][ht+1][s])*Hermitian(uy[ht][ht][s]);
	  dPlaqxz=ux[ht][ht][s]*u[ht+1][ht][s]*Hermitian(ux[ht][ht][s+1])*Hermitian(u[ht][ht][s]);
	  dPlaqyz=uy[ht][ht][s]*u[ht][ht+1][s]*Hermitian(uy[ht][ht][s+1])*Hermitian(u[ht][ht][s]);
	  out << s << "\t";
		SUNLieMatrix_OUT(out, dPlaqxy);
		out << "\t";
		SUNLieMatrix_OUT(out, dPlaqxz);
		out << "\t";
		SUNLieMatrix_OUT(out, dPlaqyz);
		out << endl << endl;
	}
	out.close();	

	// dump B's, x,z const;

	sprintf(fname,"%s/snapshot/b_y_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);	
	for (int s=1;s<=NUMT;s++) {
	  dPlaqxy=ux[ht][s][h]*uy[ht+1][s][h]*Hermitian(ux[ht][s+1][h])*Hermitian(uy[ht][s][h]);
	  dPlaqxz=ux[ht][s][h]*u[ht+1][s][h]*Hermitian(ux[ht][s][h+1])*Hermitian(u[ht][s][h]);
	  dPlaqyz=uy[ht][s][h]*u[ht][s+1][h]*Hermitian(uy[ht][s][h+1])*Hermitian(u[ht][s][h]);
	  out << s << "\t";
	    SUNLieMatrix_OUT(out, dPlaqxy);
	    out << "\t";
	    SUNLieMatrix_OUT(out, dPlaqxz);
	    out << "\t";
	    SUNLieMatrix_OUT(out, dPlaqyz);
		out << endl << endl;
	}
	out.close();	

	// dump B's, y,z const;

	sprintf(fname,"%s/snapshot/b_x_%.2f.dat",pdirname,time);
	out.open(fname, ios::out);	
	for (int s=1;s<=NUMT;s++) {
	  dPlaqxy=ux[s][ht][h]*uy[s+1][ht][h]*Hermitian(ux[s][ht+1][h])*Hermitian(uy[s][ht][h]);
	  dPlaqxz=ux[s][ht][h]*u[s+1][ht][h]*Hermitian(ux[s][ht][h+1])*Hermitian(u[s][ht][h]);
	  dPlaqyz=uy[s][ht][h]*u[s][ht+1][h]*Hermitian(uy[s][ht][h+1])*Hermitian(u[s][ht][h]);
	  out << s << "\t";
	    SUNLieMatrix_OUT(out, dPlaqxy);
	    out << "\t";
	    SUNLieMatrix_OUT(out, dPlaqxz);
	    out << "\t";
	    SUNLieMatrix_OUT(out, dPlaqyz);
		out << endl << endl;
	}
	out.close();	

	// Additional testing:


		// dump w's in z-direction
		sprintf(fname,"%s/snapshot/w_z_set_%.2f.dat",pdirname,time);
		out.open(fname, ios::out);
	 for (int i=0;i<NW;i++) {
		for (int s=0;s<=NUM+1;s++) { 
			out << s << setprecision(8) << "\t";
			out << sqrt(tr2(w[ht][ht][s][i])) << "\t";
			SUNmatrix_OUT(out, w[ht][ht][s][i]);
			out << endl;
		}
		out << endl;
		}
		out.close();

		// dump w's in y-direction
		sprintf(fname,"%s/snapshot/w_y_set_%.2f.dat",pdirname,time);
		out.open(fname, ios::out);
	 for (int i=0;i<NW;i++) {
		for (int s=0;s<=NUMT+1;s++) { 
			out << s << setprecision(8) << "\t";
			out << sqrt(tr2(w[ht][s][h][i])) << "\t";
			SUNmatrix_OUT(out, w[ht][s][h][i]);
			out << endl;
		}
		out << endl;
		}
		out.close();

		// dump w's in x-direction
		sprintf(fname,"%s/snapshot/w_x_set_%.2f.dat",pdirname,time);
		out.open(fname, ios::out);
	 for (int i=0;i<NW;i++) {
		for (int s=0;s<=NUMT+1;s++) { 
			out << s << setprecision(8) << "\t";
			out << sqrt(tr2(w[s][ht][h][i])) << "\t";
			SUNmatrix_OUT(out, w[s][ht][h][i]);
			out << endl;
		}
		out << endl;
		}
		out.close();
		
		// dump u's in z-direction
		sprintf(fname,"%s/snapshot/u_set_%.2f.dat",pdirname,time);
		out.open(fname, ios::out);
	 for (int s=0;s<=NUM+1;s++) {
			out << s << "\t";
			SUNLieMatrix_OUT(out, u[ht][ht][s]);
			out << endl;
		}
		out.close();


       	// dump uy's in z-direction
		sprintf(fname,"%s/snapshot/uy_set_%.2f.dat",pdirname,time);
		out.open(fname, ios::out);
	 for (int s=0;s<=NUM+1;s++) {
			out << s << "\t";
			SUNLieMatrix_OUT(out, uy[ht][ht][s]);
			out << endl;
		}
		out.close();

		// dump ux's in z-direction
		sprintf(fname,"%s/snapshot/ux_set_%.2f.dat",pdirname,time);
		out.open(fname, ios::out);
	 for (int s=0;s<=NUM+1;s++) {
			out << s << "\t";
			SUNLieMatrix_OUT(out, ux[ht][ht][s]);
			out << endl;
		}
		out.close();

}
