#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>

#include "sunmat.h"
#include "random.h"
#include "crossplatform.h"
#include "solve.h"
#include "outputroutines.h"
#include "outputangle.h"
#include "n20.h"
#include "filter.h"

// v2 symmetric time derivative 

using namespace std;

// these global vars are initialized from parameters file
// defaults set here are overridden by that file
// (!)MPI: All variables defined here should also appear 
//         in broadcastParametersMPI().
int    NUM=20,NUMT=20,MIMICK1D=0,STEPS=40000,UPDATE=100,SNAPUPDATE=1000,ENERGYUPDATE=1000;
int    ANGLEUPDATE=1000, ANGLEBINS1=10, ANGLEBINS2=20, ANGLEBINS3=30;
int    CHECKPOINTUPDATE=10000, TIMEOUT=3600, STATISTICSUPDATE=100;
int    USEPOLYHEDRA=1,NW=8,NPHI=5,NZ=10;
int    RANDOMSEED=19691123, RANDOMSEEDUSED=0;
int    PERTURB=0,MIMICKSUN=0,COUPLE_CURRENTS=1,FILTER_FUNC=0,INITIAL_CONDITION_TYPE=0;
// (!)MPI: If you think of changing the following floats to double,
//         remember that you have to do so also in broadcastParametersMPI().
double A=0.2,EPS=0.05,XI=10,MU2=10,SIG=0.06,PERTURBSIG=0.0,FILTER_WIDTH=1.0,MAXEVIOLATION=1;
int NUMx; // x-variable (= NUMT usually; smaller for parallel application)

// these hold the current values; note w[v] = a[v] wo[v] + b[v] wz[v]  
SUNmatrix ***pix,***piy,***ez,****w;
SUNLieMatrix ***u,***ux,***uy;

// these hold the updated values
SUNmatrix ***Pix,***Piy,***Ez,****W;
SUNLieMatrix ***U,***Ux,***Uy;

// these hold the old values of W used for making it's 
// time derivative symmetric in t
SUNmatrix ****ow;

// j holds current at t; oj holds current at t-eps
SUNmatrix ****j,****oj;

// vertex array - x=0,y=1,z=2 - and coefficient arrays - a and b
double **v,*a,*b;

//Initial Gauss law violation
SUNmatrix InitialGLv;

// InitialEnergy of the system
double InitialEnergy = 0;

// Holds the percentage difference from the initial energy
double energyDiff = 0;

// stores the running integrated value of the HL contribution to the energy
//float HLenergy = 0;
double HLenergy = 0;  // increase accuracy of integrated value.

// stores timing information (time struct as used by gettimeofday())
timeval starttime;  // time at program start
timeval outputtime; // time of last outputmeasurement()
long outputcounter; // counts number of outputmeasurements

// output files
fstream phirms_out;
fstream jrms_out;
fstream cbar_out;
fstream energy_out;
fstream gausslaw_out;
fstream clengths_out;

// initialize global arrays
// => MPI implemenation uses allocateMemoryMPI()
void allocateMemory() {

        // Notes:  Need to clean this up to call a generic allocation routine for 3 and 4 dimensional arrays

    	cout << "==> Allocating memory\n";

        // indices x,y,z
        
	pix = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) pix[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) pix[sx][sy] = new SUNmatrix[NUM+2];

	piy = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) piy[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) piy[sx][sy] = new SUNmatrix[NUM+2];

	ez = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) ez[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) ez[sx][sy] = new SUNmatrix[NUM+2];

        // indices x,y,z,W-index
        
	w = new SUNmatrix***[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) w[sx] = new SUNmatrix**[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) w[sx][sy] = new SUNmatrix*[NUM+2];
	for (int i=0;i<NUMT+2;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) w[i][j][k] = new SUNmatrix[NW];

        // indices x,y,z
        
	u = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) u[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) u[sx][sy] = new SUNLieMatrix[NUM+2];

	ux = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) ux[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) ux[sx][sy] = new SUNLieMatrix[NUM+2];

	uy = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) uy[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) uy[sx][sy] = new SUNLieMatrix[NUM+2];

        // updated values have same indices as above
        
	Pix = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Pix[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Pix[sx][sy] = new SUNmatrix[NUM+2];

	Piy = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Piy[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Piy[sx][sy] = new SUNmatrix[NUM+2];

	Ez = new SUNmatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Ez[sx] = new SUNmatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Ez[sx][sy] = new SUNmatrix[NUM+2];

	W = new SUNmatrix***[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) W[sx] = new SUNmatrix**[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) W[sx][sy] = new SUNmatrix*[NUM+2];
	for (int i=0;i<NUMT+2;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) W[i][j][k] = new SUNmatrix[NW];

	U = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) U[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) U[sx][sy] = new SUNLieMatrix[NUM+2];

	Ux = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Ux[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Ux[sx][sy] = new SUNLieMatrix[NUM+2];

	Uy = new SUNLieMatrix**[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) Uy[sx] = new SUNLieMatrix*[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) Uy[sx][sy] = new SUNLieMatrix[NUM+2];

        // indices x,y,z,W-index
	ow = new SUNmatrix***[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) ow[sx] = new SUNmatrix**[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) ow[sx][sy] = new SUNmatrix*[NUM+2];
	for (int i=0;i<NUMT+2;i++) for (int j=0;j<NUMT+2;j++) for (int k=0; k<NUM+2; k++) ow[i][j][k] = new SUNmatrix[NW];

        // indices x,y,z,(x,y,z)

	j = new SUNmatrix***[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) j[sx] = new SUNmatrix**[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) j[sx][sy] = new SUNmatrix*[NUM+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) j[sx][sy][sz] = new SUNmatrix[3];

	oj = new SUNmatrix***[NUMT+2]; 
	for (int sx=0;sx<NUMT+2;sx++) oj[sx] = new SUNmatrix**[NUMT+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) oj[sx][sy] = new SUNmatrix*[NUM+2];
	for (int sx=0;sx<NUMT+2;sx++) for (int sy=0;sy<NUMT+2;sy++) for (int sz=0; sz<NUM+2; sz++) oj[sx][sy][sz] = new SUNmatrix[3];

        // indices nw,(x,y,z)
	v = new double*[NW]; for (int i=0;i<NW;i++) v[i] = new double[3];

        // indices nw
	a = new double[NW];
	b = new double[NW];
		
	return;
}

// initialize octahedron
void initv8() {
        cout << "==> Initializing velocities for octahedron." << endl;
	int cnt=0;
	for (int i=0;i<2;i++) {
		for (int j=0;j<2;j++) {
			for (int k=0;k<2;k++) {
				v[cnt][0] = (-2*i+1) / sqrt(3.0);
				v[cnt][1] = (-2*j+1) / sqrt(3.0);
				v[cnt][2] = (-2*k+1) / sqrt(3.0);
				cnt++;
	}}}
	return;
}

// initialize disco balls
static void initDiscoBall() {
	cout << "==> Initializing vector set for discoball Nz=" << NZ << ", Nphi=" << NPHI << ", and Nw=" << NW << endl;
	double z,rho,PI=4*atan((double)1);
	int cnt=0;
	for (int i=0;i<NZ;i++) {
		z=-1.+(2.*i+1)/NZ;
		rho=sqrt(1.-z*z);
		for (int j=0;j<NPHI;j++) {
			v[cnt][0]=rho*cos(2*PI/NPHI*j);
			v[cnt][1]=rho*sin(2*PI/NPHI*j);
			v[cnt][2]=z;
			cnt++;
		}
	}
	return;
}

// initialize velocities and coefficients
void initializeVelocitiesAndCoefficients()
{
	// first initialize velocities
	if (USEPOLYHEDRA==1) {
	    // polyhedra
	    switch (NW) {
	    case 8:
	      initv8();
	      break;
	    case 20:
	      initv20();
	      break;
	    default :
	      cout << "NW=" << NW << " is an unsupported polyhedra.  Exiting now." << endl;			
	      exit(0);
	    }
	}
	else {
		// disco ball
		initDiscoBall();
	}
	// initialize a's and b's
        cout << "==> Initializing coefficients" << endl;
	double temp=0;
	for (int i=0;i<NW;i++) {
		temp = v[i][2]; temp *= XI*temp; temp += 1; temp *= temp;
		//making coefficients dimensionless
		a[i] = MU2/temp*A*A; 
		b[i] = XI*v[i][2]*a[i];
		//cout << "(" << v[i][0] << "," << v[i][1] << "," << v[i][2] << ")";
		//cout << "--> (" << a[i] << "," << b[i] << ")" << endl;
	}
}

// enforces boundary conditions by explicitly copying values into the padding elements
// => MPI implemenation uses enforceBCsforslice(int sx)
void enforceBCs()
{

//fill in z=0,NUM+1 values for main x,y
  for (int sx=1;sx<=NUMx;sx++) 
    for (int sy=1;sy<=NUMT;sy++) { 
	pix[sx][sy][0] = pix[sx][sy][NUM]; 	pix[sx][sy][NUM+1] = pix[sx][sy][1];
	piy[sx][sy][0] = piy[sx][sy][NUM]; 	piy[sx][sy][NUM+1] = piy[sx][sy][1];
	ez[sx][sy][0] = ez[sx][sy][NUM]; 	ez[sx][sy][NUM+1] = ez[sx][sy][1];
	u[sx][sy][0] = u[sx][sy][NUM]; 		u[sx][sy][NUM+1] = u[sx][sy][1];
	ux[sx][sy][0] = ux[sx][sy][NUM]; 	ux[sx][sy][NUM+1] = ux[sx][sy][1];
	uy[sx][sy][0] = uy[sx][sy][NUM]; 	uy[sx][sy][NUM+1] = uy[sx][sy][1];
	for (int i=0;i<NW;i++)  {
	  w[sx][sy][0][i] = w[sx][sy][NUM][i]; 	w[sx][sy][NUM+1][i] = w[sx][sy][1][i];
	}
    }
  
  //now fill in y=0,NUMT+1 values for main x and all z
  for (int sx=1;sx<=NUMx;sx++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      	pix[sx][0][sz] = pix[sx][NUMT][sz]; 	pix[sx][NUMT+1][sz] = pix[sx][1][sz];
	piy[sx][0][sz] = piy[sx][NUMT][sz]; 	piy[sx][NUMT+1][sz] = piy[sx][1][sz];
	ez[sx][0][sz] = ez[sx][NUMT][sz]; 	ez[sx][NUMT+1][sz] = ez[sx][1][sz];
	u[sx][0][sz] = u[sx][NUMT][sz]; 	u[sx][NUMT+1][sz] = u[sx][1][sz];
	ux[sx][0][sz] = ux[sx][NUMT][sz]; 	ux[sx][NUMT+1][sz] = ux[sx][1][sz];
	uy[sx][0][sz] = uy[sx][NUMT][sz]; 	uy[sx][NUMT+1][sz] = uy[sx][1][sz];
	for (int i=0;i<NW;i++)  {
	  w[sx][0][sz][i] = w[sx][NUMT][sz][i]; 	w[sx][NUMT+1][sz][i] = w[sx][1][sz][i];
  	}
    }
  
  //finally fill in x=0,NUMT+1 values for general y,z

  for (int sy=0;sy<=NUMT+1;sy++) 
    for (int sz=0;sz<=NUM+1;sz++) { 
      pix[0][sy][sz] = pix[NUMx][sy][sz]; 	pix[NUMx+1][sy][sz] = pix[1][sy][sz];
      piy[0][sy][sz] = piy[NUMx][sy][sz]; 	piy[NUMx+1][sy][sz] = piy[1][sy][sz];
      ez[0][sy][sz] = ez[NUMx][sy][sz]; 	ez[NUMx+1][sy][sz] = ez[1][sy][sz];
      u[0][sy][sz] = u[NUMx][sy][sz]; 		u[NUMx+1][sy][sz] = u[1][sy][sz];
      ux[0][sy][sz] = ux[NUMx][sy][sz]; 	ux[NUMx+1][sy][sz] = ux[1][sy][sz];
      uy[0][sy][sz] = uy[NUMx][sy][sz]; 	uy[NUMx+1][sy][sz] = uy[1][sy][sz];
      for (int i=0;i<NW;i++)  {
	w[0][sy][sz][i] = w[NUMx][sy][sz][i]; 	w[NUMx+1][sy][sz][i] = w[1][sy][sz][i];
      }
    }
}

/*
 * copies Pi,Phi,Ez,U,W to pi,phi,ez,u,w and then calls enforceBCs()
 *
*/
void copyDown() {
for (int sx=1;sx<=NUMx;sx++) 
    for (int sy=1;sy<=NUMT;sy++)
      for (int sz=1; sz<=NUM;sz++) { 
	pix[sx][sy][sz] = Pix[sx][sy][sz];
	piy[sx][sy][sz] = Piy[sx][sy][sz];
	ez[sx][sy][sz] = Ez[sx][sy][sz];
	u[sx][sy][sz] = U[sx][sy][sz];
	ux[sx][sy][sz] = Ux[sx][sy][sz];
	uy[sx][sy][sz] = Uy[sx][sy][sz];
	// don't forget to save w in ow
	for (int i=0;i<NW;i++)  ow[sx][sy][sz][i]=w[sx][sy][sz][i];
	for (int i=0;i<NW;i++)  w[sx][sy][sz][i]=W[sx][sy][sz][i];
      }
enforceBCs();
}

SUNmatrix generateRandomSUNmatrix(double sig) {
	
	SUNmatrix temp;
	
  #if defined SU2 || defined SU2double
			switch(MIMICKSUN)
			{
			case 1:  // U(1)
				temp.a = randGauss(sig)*A*A*A;
				temp.b = 0;
				temp.c = 0;
			        break;
			case 0:  // default = SU(2)
			case 2:  // SU(2)
			default:
				temp.a = randGauss(sig)*A*A*A;
				temp.b = randGauss(sig)*A*A*A;
				temp.c = randGauss(sig)*A*A*A;
				break;
			}
  #elif defined SU3 || defined SU3double || defined SU3tvmet
			double a1,a2,a3,a4,a5,a6,a7,a8;
			switch(MIMICKSUN)
			{
			case 1:  // U(1)
				a1 = randGauss(sig)*A*A*A;
				a2 = 0;
				a3 = 0;
				a4 = 0;
				a5 = 0;
				a6 = 0;
				a7 = 0;
				a8 = 0;
				break;
			case 2:  // SU(2)
				a1 = randGauss(sig)*A*A*A;
				a2 = randGauss(sig)*A*A*A;
				a3 = randGauss(sig)*A*A*A;
				a4 = 0;
				a5 = 0;
				a6 = 0;
				a7 = 0;
				a8 = 0;
				break;
			case 0:  // default = SU(3)
			case 3:  // SU(3)
			default:  // anything else
				a1 = randGauss(sig)*A*A*A;
				a2 = randGauss(sig)*A*A*A;
				a3 = randGauss(sig)*A*A*A;
				a4 = randGauss(sig)*A*A*A;
				a5 = randGauss(sig)*A*A*A;
				a6 = randGauss(sig)*A*A*A;
				a7 = randGauss(sig)*A*A*A;
				a8 = randGauss(sig)*A*A*A;
			}
			temp = SUNmatrix(a1,a2,a3,a4,a5,a6,a7,a8);
	#else
	#error Please provide your SU(N) code here.
	#endif

	return temp;
}	

// Setting of initial conditions for one slice
// (split into slices because of MPI implementation)
void setInitialConditionsForSlice(int sx)
{
	
	#if defined SU2 || defined SU2double
		const double sig=SIG/sqrt(A)/A; // standard deviation for SU(2)
	#elif defined SU3 || defined SU3double
		const double sig=SIG/sqrt(A)/A; // standard deviation for SU(3)
	#elif defined SU3tvmet
		const double sig=SIG/sqrt(A)/A; // standard deviation for SU(3)
		SU3matrix::initGellMannMatrices();
	#else
	#error Please provide your SU(N) code here.
	#endif

	SUNmatrix temp;

  switch(INITIAL_CONDITION_TYPE){
	  case 0:
	  	//cout << "--- Initializing W fields" << endl;
			for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++)
			{
				// respects GL, gives nonzero flux
				temp.reset();		
	  		for (int j=0;j<NW;j++)
				{
				    //generic but gives nonzero T^0i
				    //should give energy conservation
				
				    w[sx][sy][sz][j] = generateRandomSUNmatrix(sig);
				
				    if (MIMICK1D==1)
				    {
							//since then the W's are more correlated
							w[sx][sy][sz][j]*=A;
							w[sx][sy][sz][j]=w[1][1][sz][j];
				    }
				    temp+=w[sx][sy][sz][j];
				}
				// guarantee that there is no gauss-violation and
				// load w at t-eps
				for (int j=0;j<NW;j++)
				{
				    w[sx][sy][sz][j]-=temp/NW;
				    ow[sx][sy][sz][j]=w[sx][sy][sz][j];
				}
		
				#if defined SU2 || defined SU2double
						u[sx][sy][sz].reset();
						ux[sx][sy][sz].reset();
						uy[sx][sy][sz].reset();
						u[sx][sy][sz].a = 1;
						ux[sx][sy][sz].a =1;
						uy[sx][sy][sz].a =1;
				#elif defined SU3 || defined SU3double || defined SU3tvmet
						u[sx][sy][sz] = SUNLieMatrix(1,0,0,0,0,0,0,0,0); 
						ux[sx][sy][sz] = SUNLieMatrix(1,0,0,0,0,0,0,0,0); 
						uy[sx][sy][sz] = SUNLieMatrix(1,0,0,0,0,0,0,0,0); 
				#else
				#error Please provide your SU(N) code here.
				#endif
			}
			break;
		case 1:
	  	//cout << "--- Initializing A fields" << endl;
	  	for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++) {
			  for (int j=0;j<NW;j++) {
			    w[sx][sy][sz][j].reset();
			    ow[sx][sy][sz][j].reset();
			  }			  
			  pix[sx][sy][sz].reset();
			  piy[sx][sy][sz].reset();
			  ez[sx][sy][sz].reset();
			  u[sx][sy][sz] = mexpi(-1.0*generateRandomSUNmatrix(sig));
			  ux[sx][sy][sz] = mexpi(-1.0*generateRandomSUNmatrix(sig));
			  uy[sx][sy][sz] = mexpi(-1.0*generateRandomSUNmatrix(sig));
			  if (MIMICK1D==1)
			  {
			  	u[sx][sy][sz] = u[1][1][sz];
			  	ux[sx][sy][sz] = ux[1][1][sz];
			  	uy[sx][sy][sz] = uy[1][1][sz];
			  }
			}
			break;
		case 2:
			for (int sy = 1; sy <= NUM; sy++)
				for (int sz = 1; sz <= NUM; sz++)	 {
					pix[sx][sy][sz].a = A*sig*cos(2*M_PI*(sz-1)*1/NUM);
					pix[sx][sy][sz].b = A*sig*cos(2*M_PI*(sz-1)*1/NUM+1);
					pix[sx][sy][sz].c = A*sig*cos(2*M_PI*(sz-1)*1/NUM+2);
					Pix[sx][sy][sz] = pix[sx][sy][sz];
				}
			for (int sy = 1; sy <= NUM; sy++)
				for (int sz = 1; sz <= NUM; sz++)	 {
					piy[sx][sy][sz].a = A*sig*cos(2*M_PI*(sz-1)*1/NUM+0.5);
					piy[sx][sy][sz].b = A*sig*cos(2*M_PI*(sz-1)*1/NUM+1.5);
					piy[sx][sy][sz].c = A*sig*cos(2*M_PI*(sz-1)*1/NUM+2.5);
					Piy[sx][sy][sz] = piy[sx][sy][sz];
				}
			for (int sy = 1; sy <= NUMT; sy++)
				for (int sz = 1; sz <= NUM; sz++)	 {
					ux[sx][sy][sz].reset();
					uy[sx][sy][sz].reset();
					u[sx][sy][sz].reset();
					ux[sx][sy][sz].a = 1.0;
					uy[sx][sy][sz].a = 1.0;
					u[sx][sy][sz].a = 1.0;
					Ux[sx][sy][sz] = ux[sx][sy][sz];
					Uy[sx][sy][sz] = uy[sx][sy][sz];
					U[sx][sy][sz] = u[sx][sy][sz];
					for (int i = 0; i < NW; i++) {
							w[sx][sy][sz][i].reset();
							W[sx][sy][sz][i].reset();
					}
				}
			break;
	default:
	  	cout << "==> Unknown INITIAL_CONDITION_TYPE" << endl;
	  	cout << "==> Aborting" << endl;
	  	exit(0);
  }
  	
}

// initializes the variables and then calls enforceBCs()
// => MPI implemenation uses setInitialConditionsMPI()
//    and setInitialConditionsForSlice(int sx)
void setInitialConditions()
{
  cout << "==> Initializing variables\n";

	srand(RANDOMSEEDUSED);

	for (int sx=1;sx<=NUMx;sx++) 
	{
		setInitialConditionsForSlice(sx);
	};

	enforceBCs();
  
}

void perturbInitialConditions()
{
  cout << "==> Perturbing initial condition\n";

  // always use time-based seed here
  srand((unsigned int)time(0));

	for (int sx=1;sx<=NUMx;sx++) 
	{
		perturbInitialConditionsForSlice(sx);
	};

	enforceBCs();

}

void perturbInitialConditionsForSlice(int sx)
{
#if defined SU2 || defined SU2double
	const double sig=PERTURBSIG/sqrt(A)/A; // standard deviation for SU(2)
#elif defined SU3 || defined SU3double
	const double sig=PERTURBSIG/sqrt(A)/A; // standard deviation for SU(3)
#elif defined SU3tvmet
	const double sig=PERTURBSIG/sqrt(A)/A; // standard deviation for SU(3)
	SU3matrix::initGellMannMatrices();
#else
#error Please provide your SU(N) code here.
#endif

	SUNmatrix temp;
	SUNLieMatrix id;
	
	#if defined SU2 || defined SU2double
		id.reset();
	  id.a = 1;
	#elif defined SU3 || defined SU3double || defined SU3tvmet
	  id = SUNLieMatrix(1,0,0,0,0,0,0,0,0); 
	#else
  #error Please provide your SU(N) code here.
  #endif

  switch(INITIAL_CONDITION_TYPE){
	  case 0:
			for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++)
			{
				// respects GL, gives nonzero flux
				temp.reset();
		
				for (int j=0;j<NW;j++)
				{
				    //generic but gives nonzero T^0i
				    //should give energy conservation
		
				    if (MIMICK1D==1)
				    {
						  //since then the W's are more correlated
						  w[sx][sy][sz][j] += A*generateRandomSUNmatrix(sig);
						  w[sx][sy][sz][j]=w[1][1][sz][j];
				    }
				    else {
				      w[sx][sy][sz][j] += generateRandomSUNmatrix(sig);
				    }
				  
				    temp += w[sx][sy][sz][j];
				}
				// guarantee that there is no gauss-violation and
				// load w at t-eps
				for (int j=0;j<NW;j++)
				{
				    w[sx][sy][sz][j]-=temp/NW;
				    ow[sx][sy][sz][j]=w[sx][sy][sz][j];
				}
		
			}
			break;
	  case 1:
	  	for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++) {
			  u[sx][sy][sz] += mexpi(-1.0*generateRandomSUNmatrix(sig));
			  ux[sx][sy][sz] += mexpi(-1.0*generateRandomSUNmatrix(sig));
			  uy[sx][sy][sz] += mexpi(-1.0*generateRandomSUNmatrix(sig));
			  u[sx][sy][sz].normalize();
			  ux[sx][sy][sz].normalize();
			  uy[sx][sy][sz].normalize();
			}	  	
	  	break;
	  default:
	  	break;
	  }
	  return;
}

// loads current j - not very efficient should fold into doInc
void loadjs() {
// remember:  j[NUMT+2][NUMT+2][NUM+2][3], v[NW][3], w[NUMT+2][NUMT+2][NUM+2][NW]
  for (int sx=0;sx<=NUMx+1;sx++)
    for (int sy=0;sy<=NUMT+1;sy++)
      for (int sz=0;sz<=NUM+1;sz++)
	for (int i=0;i<3;i++) {
	  //j[sx][sy][sz][i].a = 0; j[sx][sy][sz][i].b = 0; j[sx][sy][sz][i].c = 0;
	  j[sx][sy][sz][i].reset();
	  if (COUPLE_CURRENTS) {
		  for (int k=0;k<NW;k++) {
		    j[sx][sy][sz][i] += w[sx][sy][sz][k]*v[k][i];
		  }
		  j[sx][sy][sz][i] /= NW;
	  }
	} 
  return;
}

// save current j for later use
void savejs() {
  for (int sx=0;sx<=NUMx+1;sx++)
    for (int sy=0;sy<=NUMT+1;sy++)
      for (int sz=0;sz<=NUM+1;sz++)
	  for (int i=0;i<3;i++) oj[sx][sy][sz][i] = j[sx][sy][sz][i];
}

// update HLEnergy
// uses midpoint rule to compute integral in def of HL energy
void updateHLenergy(double eps) {
	double res=0, resy, resz;
	// note: this routine assumes that j and oj have already been "loaded"
	// remember:  J[NUMT+2]][NUMT+2][NUM+2][3]
	if (ACCURATESUMMATION)
	{
		for (int sx=1;sx<=NUMx;sx++)
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1;sz<=NUM;sz++)
				{		
					resz -= tr(pix[sx][sy][sz],oj[sx][sy][sz][0]+j[sx][sy][sz][0]+transportRight(oj[sx+1][sy][sz][0]+j[sx+1][sy][sz][0],ux[sx][sy][sz]))/2;
					resz -= tr(piy[sx][sy][sz],oj[sx][sy][sz][1]+j[sx][sy][sz][1]+transportRight(oj[sx][sy+1][sz][1]+j[sx][sy+1][sz][1],uy[sx][sy][sz]))/2;
					resz += tr(ez[sx][sy][sz],oj[sx][sy][sz][2]+j[sx][sy][sz][2]+transportRight(oj[sx][sy][sz+1][2]+j[sx][sy][sz+1][2],u[sx][sy][sz]))/2;

					// Test:
					//resz -= tr(pix[sx][sy][sz],oj[sx][sy][sz][0]+j[sx][sy][sz][0]+transportRight(oj[sx+1][sy][sz][0]+j[sx+1][sy][sz][0],ux[sx][sy][sz]))/2;
					//resz -= tr(piy[sx][sy][sz],oj[sx][sy][sz][1]+j[sx][sy][sz][1]+transportRight(oj[sx][sy+1][sz][1]+j[sx][sy+1][sz][1],uy[sx][sy][sz]))/2;
					//resz += tr(ez[sx][sy][sz],oj[sx][sy][sz][2]+j[sx][sy][sz][2]+transportRight(oj[sx][sy][sz+1][2]+j[sx][sy][sz+1][2],u[sx][sy][sz]))/2;
				};
				resy += resz;
			};
			res += resy;
		};
	}
	else
	{
		// old routine
		for (int sx=1;sx<=NUMx;sx++)
		  for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1;sz<=NUM;sz++)
			  {		
			res -= tr(pix[sx][sy][sz],oj[sx][sy][sz][0]+j[sx][sy][sz][0]+transportRight(oj[sx+1][sy][sz][0]+j[sx+1][sy][sz][0],ux[sx][sy][sz]))/2;
			res -= tr(piy[sx][sy][sz],oj[sx][sy][sz][1]+j[sx][sy][sz][1]+transportRight(oj[sx][sy+1][sz][1]+j[sx][sy+1][sz][1],uy[sx][sy][sz]))/2;
			res += tr(ez[sx][sy][sz],oj[sx][sy][sz][2]+j[sx][sy][sz][2]+transportRight(oj[sx][sy][sz+1][2]+j[sx][sy][sz+1][2],u[sx][sy][sz]))/2;
			  }
	};
    HLenergy += eps/A*res;
}

// computes energy in longitudinal electric fields
double ezEnergyDensity() {
        // E's and Pi's are defined at half-integer times and the fields 
        // at integer times so in order to most accurately compute the energy density
        // we should average the values of E and Pi at t-eps/2 and t+eps/2
        // to obtain a better approximation to their value at t
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;

		for (int sx=1;sx<=NUMx;sx++)
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1;sz<=NUM;sz++)
				{
					resz += tr2(ez[sx][sy][sz]+Ez[sx][sy][sz])/4;
				};
				resy += resz;
			};
			resd += resy;
		};
		resd/=NUMT*A;
		resd/=NUMT*A;
		resd/=NUM*A;	
		return resd/A;
	}
	else
	{
		double res=0;

		for (int sx=1;sx<=NUMx;sx++)
		  for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1;sz<=NUM;sz++)
			  {
			res += tr2(ez[sx][sy][sz]+Ez[sx][sy][sz])/4;
			  }
		res/=NUMT*A;
		res/=NUMT*A;
		res/=NUM*A;	
		return res/A;
	};
}

// computes energy in transverse electric fields
double etEnergyDensity() {
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;
			// E's and Pi's are defined at half-integer times and the fields 
			// at integer times so in order to most accurately compute the energy density
			// we should average the values of E and Pi at t-eps/2 and t+eps/2
			// to obtain a better approximation to their value at t
		for (int sx=1;sx<=NUMx;sx++)
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1;sz<=NUM;sz++)
				{
					resz += (tr2(pix[sx][sy][sz]+Pix[sx][sy][sz]))/4;
					resz += (tr2(piy[sx][sy][sz]+Piy[sx][sy][sz]))/4;
				};
				resy += resz;
			};
			resd += resy;
		}
		resd/=NUMT*A;
		resd/=NUMT*A;
		resd/=NUM*A;
		return resd/A;

	}
	else
	{
		double res=0;
			// E's and Pi's are defined at half-integer times and the fields 
			// at integer times so in order to most accurately compute the energy density
			// we should average the values of E and Pi at t-eps/2 and t+eps/2
			// to obtain a better approximation to their value at t
		for (int sx=1;sx<=NUMx;sx++)
		  for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1;sz<=NUM;sz++)
			  {
			res += (tr2(pix[sx][sy][sz]+Pix[sx][sy][sz]))/4;
			res += (tr2(piy[sx][sy][sz]+Piy[sx][sy][sz]))/4;
			  }
		res/=NUMT*A;
		res/=NUMT*A;
		res/=NUM*A;
		return res/A;
	}
}

inline double PlaquetteAction(SUNLieMatrix dPlaq)
{
	int N = SUNLieMatrix::N(); // 2 for SU(2); 3 for SU(3)
	//return (2*N)*(1. - tr(dPlaq)/N);
	return 2*((double)N - tr(dPlaq));
}

SUNLieMatrix dPlaqXY(int sx, int sy, int sz)
{
	return ux[sx][sy][sz]*uy[sx+1][sy][sz]*Hermitian(ux[sx][sy+1][sz])*Hermitian(uy[sx][sy][sz]);
}

SUNLieMatrix dPlaqXZ(int sx, int sy, int sz)
{
	return ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
}

SUNLieMatrix dPlaqYZ(int sx, int sy, int sz)
{
	return uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
}

// computes the energy in longitudinal magnetic fields
double bzEnergyDensity() {
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;

		SUNLieMatrix dPlaqxy;
		for (int sx=1;sx<=NUMx;sx++)
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1;sz<=NUM;sz++)
				{
					// these guys are defined at integer times so no need to average here
					resz += PlaquetteAction(dPlaqXY(sx,sy,sz));

				}
				resy += resz;
			};
			resd += resy;
		};
		resd/=NUMT*A;
		resd/=NUMT*A;
		resd/=NUM*A;
		return resd/A;
	}
	else{
		double res=0;

		SUNLieMatrix dPlaqxy;
		for (int sx=1;sx<=NUMx;sx++)
		for (int sy=1;sy<=NUMT;sy++)
		for (int sz=1;sz<=NUM;sz++)
		{
			// these guys are defined at integer times so no need average here
			res += PlaquetteAction(dPlaqXY(sx,sy,sz));
		} // for
		res/=NUMT*A;
		res/=NUMT*A;
		res/=NUM*A;
		return res/A;
	}
}

// computes the energy in transverse magnetic fields
double btEnergyDensity() {
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;
		
		for (int sx=1;sx<=NUMx;sx++)
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1;sz<=NUM;sz++)
				{
					// these guys are defined at integer times so no need to average here
					resz += PlaquetteAction(dPlaqXZ(sx,sy,sz));
					resz += PlaquetteAction(dPlaqYZ(sx,sy,sz));
				}
				resy += resz;
			}
			resd += resy;
		}
		resd/=NUMT*A;
		resd/=NUMT*A;
		resd/=NUM*A;
		return resd/A;
	}
	else
	{
		double res=0;
		SUNLieMatrix dPlaqxz,dPlaqyz;
		
		for (int sx=1;sx<=NUMx;sx++)
		for (int sy=1;sy<=NUMT;sy++)
	  for (int sz=1;sz<=NUM;sz++)
		{
      // these guys are defined at integer times so no need average here
			res += PlaquetteAction(dPlaqXZ(sx,sy,sz));
			res += PlaquetteAction(dPlaqYZ(sx,sy,sz));
		}
		res/=NUMT*A;
		res/=NUMT*A;
		res/=NUM*A;
		return res/A;

	}
}

//initial energy density

double initialED()
{
	switch(INITIAL_CONDITION_TYPE)
	{
		case 0:
			// w fields
			if (ACCURATESUMMATION)
			{
				double resd=0, resy, resz, resw;
					for (int sx=1;sx<=NUMx;sx++)
					{
						resy = 0;
						for (int sy=1;sy<=NUMT;sy++)
						{
							resz = 0;
							for (int sz=1;sz<=NUM;sz++)
							{
								resw = 0;
								for (int i=0;i<NW;i++)
								{
									resw+=tr2(w[sx][sy][sz][i])/a[i];
								};
								resz += resw;
							};
							resy += resz;
						};
						resd += resy;
					};
					resd/=NW*2;
					return resd/NUMT/NUMT/NUM;
			}
			else
			{
			  double res=0;
			  for (int sx=1;sx<=NUMx;sx++)
				for (int sy=1;sy<=NUMT;sy++)
				  for (int sz=1;sz<=NUM;sz++)
					for (int i=0;i<NW;i++)
				  res+=tr2(w[sx][sy][sz][i])/a[i];   
			  res/=NW*2;
			  return res/NUMT/NUMT/NUM;
			}
			break;
		case 1:
			return energyDensity();
		default:
			return 0;
	}
}

// a convenience method for getting total energyDensity
double energyDensity() {
	if (ACCURATESUMMATION)
	{
		double res = 0;
		res += etEnergyDensity();
		res += ezEnergyDensity();
		res += bzEnergyDensity();
		res += btEnergyDensity();
		if (COUPLE_CURRENTS) res += HLenergy/NUMT/NUMT/NUM/A/A/A/A;
		return res;
	}
	else
	{
		double res = 0;
		res += etEnergyDensity();
		res += ezEnergyDensity();
		res += bzEnergyDensity();
		res += btEnergyDensity();
		if (COUPLE_CURRENTS) res += HLenergy/NUMT/NUMT/NUM/A/A/A/A;
		return res;
	}
}

// checkGaussLaw has to be split into part1 and part2 for MPI implementation
// needs: w, ow, ez, u, ux, uy, pix, piy
SUNmatrix checkGaussLawPart1() {
	SUNmatrix temp, tempy, tempz;
	temp.reset();
	if (ACCURATESUMMATION)
	{
		for (int sx=1;sx<=NUMx;sx++)
		{
			tempy.reset();
			for (int sy=1; sy<=NUMT;sy++)
			{
				tempz.reset();
				for (int sz=1; sz<=NUM; sz++)
				{
					if (COUPLE_CURRENTS) {
						for (int i=0;i<NW;i++) tempz += (w[sx][sy][sz][i]+ow[sx][sy][sz][i])/(2*NW);
					}
					tempz+=transportLeft(ez[sx][sy][sz-1],u[sx][sy][sz-1])-ez[sx][sy][sz]-(transportLeft(pix[sx-1][sy][sz],ux[sx-1][sy][sz])-pix[sx][sy][sz])-(transportLeft(piy[sx][sy-1][sz],uy[sx][sy-1][sz])-piy[sx][sy][sz]);
				};
				tempy+=tempz;
			};
			temp+=tempy;
		};
	}
	else
	{
		// Old routine:
		for (int sx=1;sx<=NUMx;sx++) 
		for (int sy=1; sy<=NUMT;sy++)
		for (int sz=1; sz<=NUM; sz++)
		{
		  if (COUPLE_CURRENTS) {
		  	for (int i=0;i<NW;i++) temp += (w[sx][sy][sz][i]+ow[sx][sy][sz][i])/(2*NW);
		  }
			temp+=transportLeft(ez[sx][sy][sz-1],u[sx][sy][sz-1])-ez[sx][sy][sz]-(transportLeft(pix[sx-1][sy][sz],ux[sx-1][sy][sz])-pix[sx][sy][sz])-(transportLeft(piy[sx][sy-1][sz],uy[sx][sy-1][sz])-piy[sx][sy][sz]);		
		}
	};
	return(temp);
}

double checkGaussLawPart2(SUNmatrix temp) {
	if (ACCURATESUMMATION)
	{
		double res=0;
		res=tr2(temp);
		res/=NUMT*A*A;
		res/=NUMT*A*A;
		res/=NUM*A*A;
		return sqrt(res);
	}
	else
	{
		double res=0;
		res=tr2(temp);
		res/=NUMT*A*A;
		res/=NUMT*A*A;
		res/=NUM*A*A;
		return sqrt(res);
	}
}

// compatibility function
double checkGaussLaw() {
	return checkGaussLawPart2(checkGaussLawPart1());
};

/*
float checkGaussLaw_original() {
	float res=0;
	SUNmatrix temp;
	temp.a=0; temp.b=0; temp.c=0;
	for (int sx=1;sx<=NUMx;sx++) 
	  for (int sy=1; sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM; sz++)
	      {
		//for (int i=0;i<NW;i++) temp += w[sx][sy][sz][i]/NW;
		for (int i=0;i<NW;i++) temp += (w[sx][sy][sz][i]+ow[sx][sy][sz][i])/(2*NW);
		//subtract initial GLviolation?
		temp+=transportLeft(ez[sx][sy][sz-1],u[sx][sy][sz-1])-ez[sx][sy][sz]-(transportLeft(pix[sx-1][sy][sz],ux[sx-1][sy][sz])-pix[sx][sy][sz])-(transportLeft(piy[sx][sy-1][sz],uy[sx][sy-1][sz])-piy[sx][sy][sz]);		
	}
	res=tr2(temp);
	res/=NUMT*A*A;
	res/=NUMT*A*A;
	res/=NUM*A*A;
	return sqrt(res);
}
*/

// computes j-based Cxy
double newCj(int sx, int sy, int sz) {
	// note j[NUMT+2][3]
	// note that this is not well-defined if j's are vanishing
  return sqrt(tr2(icommute(j[sx][sy][sz][0],j[sx][sy][sz][1])))/(sqrt(tr2(j[sx][sy][sz][0]))*sqrt(tr2(j[sx][sy][sz][1])));
}


// computes j-based Cxz
double Cjxz(int sx, int sy, int sz) {
	// note j[NUMT+2][3]
	// note that this is not well-defined if j's are vanishing
  return sqrt(tr2(icommute(j[sx][sy][sz][0],j[sx][sy][sz][2])))/(sqrt(tr2(j[sx][sy][sz][0]))*sqrt(tr2(j[sx][sy][sz][2])));
}

double Cjyz(int sx, int sy, int sz) {
	// note j[NUMT+2][3]
	// note that this is not well-defined if j's are vanishing
  return sqrt(tr2(icommute(j[sx][sy][sz][1],j[sx][sy][sz][2])))/(sqrt(tr2(j[sx][sy][sz][1]))*sqrt(tr2(j[sx][sy][sz][2])));
}


/* computes j-based Cbar
float Cjbar() {
	float res=0;
	for (int sx=1;sx<=NUMT;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	      res += Cj(sx,sy,sz); }
	return res/NUMT/NUMT/NUM;
}
*/

// computes j-based Cbar
double newCjbar() {
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;
		for (int sx=1;sx<=NUMx;sx++) 
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1; sz<=NUM;sz++) 
				{
					resz += newCj(sx,sy,sz); 
				};
				resy += resz;
			};
			resd += resy;
		}
		return resd/NUMT/NUMT/NUM;
	}
	else
	{
		double res=0;
		for (int sx=1;sx<=NUMx;sx++) 
		  for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++) { 
			  res += newCj(sx,sy,sz); }
		return res/NUMT/NUMT/NUM;
	}
}

// computes j-based Cbar
double Cjxzbar() {
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;
		for (int sx=1;sx<=NUMx;sx++) 
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1; sz<=NUM;sz++) 
				{
					resz += Cjxz(sx,sy,sz);; 
				};
				resy += resz;
			};
			resd += resy;
		}
		return resd/NUMT/NUMT/NUM;
	}
	else
	{
		double res=0;
		for (int sx=1;sx<=NUMx;sx++) 
		  for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++) { 
			  res += Cjxz(sx,sy,sz); }
		return res/NUMT/NUMT/NUM;
	}
}

// computes j-based Cbar
double Cjyzbar() {
	if (ACCURATESUMMATION)
	{
		double resd=0, resy, resz;
		for (int sx=1;sx<=NUMx;sx++) 
		{
			resy = 0;
			for (int sy=1;sy<=NUMT;sy++)
			{
				resz = 0;
				for (int sz=1; sz<=NUM;sz++) 
				{
					resz += Cjyz(sx,sy,sz);; 
				};
				resy += resz;
			};
			resd += resy;
		}
		return resd/NUMT/NUMT/NUM;
	}
	else
	{
		double res=0;
		for (int sx=1;sx<=NUMx;sx++) 
		  for (int sy=1;sy<=NUMT;sy++)
			for (int sz=1; sz<=NUM;sz++) { 
			  res += Cjyz(sx,sy,sz); }
		return res/NUMT/NUMT/NUM;
	}
}

// Unfortunately, inline not possible...
// inline 

// doInc() is split up 
// into part1 (loadjs: w => j
//             calculate u, ux, uy, ez, pix, piy, ez, j
//                       => j, Pix, Piy, Ez, U, Ux, Uy)
//  and part2 (calculate u, ux, uy, U, Ux, Uy, pix, piy, ez, Pix, Piy, Ez, ow 
//                       => W)
void doIncPart1(double eps) {
  //SUNmatrix c,d;
  double step=eps/A;
  SUNmatrix temp;
  SUNLieMatrix Ua,Ub,Uc,Ud;
  SUNLieMatrix dPlaqxz; //,dPlaqyz;
//  SUNmatrix tempFxz,tempFyz;


  loadjs();

/*
  cout << "-----------------------------------------------------------------------------------------------" << endl;
	for (int sx = 1; sx <= NUMT; sx++)
		for (int sy = 1; sy <= NUMT; sy++)
			for (int sz = 1; sz <= NUM; sz++)	 
				cout << setprecision (25) << "DEBUG pix - " << sx << ", " << sy << ", " << sz << " - " << pix[sx][sy][sz].a << ", " << pix[sx][sy][sz].b << ", " << pix[sx][sy][sz].c << endl;
  cout << "-----------------------------------------------------------------------------------------------" << endl;
  cout << "-----------------------------------------------------------------------------------------------" << endl;
	for (int sx = 1; sx <= NUMT; sx++)
		for (int sy = 1; sy <= NUMT; sy++)
			for (int sz = 1; sz <= NUM; sz++)	 
				cout << setprecision (25) << "DEBUG piy - " << sx << ", " << sy << ", " << sz << " - " << piy[sx][sy][sz].a << ", " << piy[sx][sy][sz].b << ", " << piy[sx][sy][sz].c << endl;
  cout << "-----------------------------------------------------------------------------------------------" << endl;
  cout << "-----------------------------------------------------------------------------------------------" << endl;
	for (int sx = 1; sx <= NUMT; sx++)
		for (int sy = 1; sy <= NUMT; sy++)
			for (int sz = 1; sz <= NUM; sz++)	 
				cout << setprecision (25) << "DEBUG piz - " << sx << ", " << sy << ", " << sz << " - " << ez[sx][sy][sz].a << ", " << ez[sx][sy][sz].b << ", " << ez[sx][sy][sz].c << endl;
  cout << "-----------------------------------------------------------------------------------------------" << endl;
*/

  for (int sx=1;sx<=NUMx;sx++) 
	  for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
		  //cout << "doInc(): sx" << sx << ",";
		  //cout << "doInc(): NUMx" << NUMx << ",";

	      // Update Ex 

	      Ua=ux[sx][sy][sz]*uy[sx+1][sy][sz]*Hermitian(ux[sx][sy+1][sz])*Hermitian(uy[sx][sy][sz]);
	      Ub=ux[sx][sy][sz]*Hermitian(uy[sx+1][sy-1][sz])*Hermitian(ux[sx][sy-1][sz])*uy[sx][sy-1][sz];
	      Uc=ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
		  Ud=ux[sx][sy][sz]*Hermitian(u[sx+1][sy][sz-1])*Hermitian(ux[sx][sy][sz-1])*u[sx][sy][sz-1];
	      dPlaqxz  = Uc;

		  temp = toAlgebraElement(Ua + Ub + Uc + Ud);

		  //cout << setprecision (20) << "DEBUG Ex : " << temp.a << ", " << temp.b << ", " << temp.c << endl;
		  
	      Pix[sx][sy][sz] = pix[sx][sy][sz] + step*(2*temp +j[sx][sy][sz][0]+transportRight(j[sx+1][sy][sz][0],ux[sx][sy][sz]))/2;

			/*
			cout << "-----------------------------------------------------------------------------------------------" << endl;
			cout << "Ua : " << Ua.a << ", " << Ua.b << ", " << Ua.c << ", " << Ua.d << endl;
			cout << "Ub : " << Ub.a << ", " << Ub.b << ", " << Ub.c << ", " << Ub.d << endl;
			cout << "Uc : " << Uc.a << ", " << Uc.b << ", " << Uc.c << ", " << Uc.d << endl;
			cout << "Ud : " << Ud.a << ", " << Ud.b << ", " << Ud.c << ", " << Ud.d << endl;
			cout << "temp : " << temp.a << ", " << temp.b << ", " << temp.c << endl;
			cout << "-----------------------------------------------------------------------------------------------" << endl;
			*/
			
	      // Update Ey 

	      Ua=Hermitian(Ua);
	      Ub=uy[sx][sy][sz]*Hermitian(ux[sx-1][sy+1][sz])*Hermitian(uy[sx-1][sy][sz])*ux[sx-1][sy][sz];
	      Uc=uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
	      Ud=uy[sx][sy][sz]*Hermitian(u[sx][sy+1][sz-1])*Hermitian(uy[sx][sy][sz-1])*u[sx][sy][sz-1];

//	      dPlaqyz = Uc;

		  temp = toAlgebraElement(Ua + Ub + Uc + Ud);
/*
		  //cout << setprecision (20) << "DEBUG Ey Ua : " << Ua.a << ", " << Ua.b << ", " << Ua.c << ", " << Ua.d << endl;
		  //cout << setprecision (20) << "DEBUG Ey Ub : " << Ub.a << ", " << Ub.b << ", " << Ub.c << ", " << Ub.d << endl;
		  //cout << setprecision (20) << "DEBUG Ey Uc : " << Uc.a << ", " << Uc.b << ", " << Uc.c << ", " << Uc.d << endl;
		  //cout << setprecision (20) << "DEBUG Ey Ud : " << Ud.a << ", " << Ud.b << ", " << Ud.c << ", " << Ud.d << endl;
*/
		  //cout << setprecision (20) << "DEBUG Ey : " << temp.a << ", " << temp.b << ", " << temp.c << endl;

	      Piy[sx][sy][sz] = piy[sx][sy][sz] + step*(2*temp +j[sx][sy][sz][1]+transportRight(j[sx][sy+1][sz][1],uy[sx][sy][sz]))/2;

	      // Update Ez 

	      Ua=Hermitian(dPlaqxz);
	      Ub=u[sx][sy][sz]*Hermitian(ux[sx-1][sy][sz+1])*Hermitian(u[sx-1][sy][sz])*ux[sx-1][sy][sz];
	      Uc=Hermitian(Uc);
	      Ud=u[sx][sy][sz]*Hermitian(uy[sx][sy-1][sz+1])*Hermitian(u[sx][sy-1][sz])*uy[sx][sy-1][sz];	

		  temp = toAlgebraElement(Ua + Ub + Uc + Ud);

		  //cout << setprecision (20) << "DEBUG Ez : " << temp.a << ", " << temp.b << ", " << temp.c << endl;

	      Ez[sx][sy][sz] = ez[sx][sy][sz] - step*(2*temp+j[sx][sy][sz][2]+transportRight(j[sx][sy][sz+1][2],u[sx][sy][sz]))/2;

	      // then use the values just obtained to update U and W

	      U[sx][sy][sz] = mexpi(step*Ez[sx][sy][sz])*u[sx][sy][sz]; 
	      Ux[sx][sy][sz] = mexpi(-1*step*Pix[sx][sy][sz])*ux[sx][sy][sz];
	      Uy[sx][sy][sz] = mexpi(-1*step*Piy[sx][sy][sz])*uy[sx][sy][sz];

	      // Take care of points on the boundary by imposing boundary condition where necessary 
		  // boundary conditions in x-direction are taken care of by 
		  // the parallelization.
		
		  if (NUMx > 1) // NUMx == 1 is the parallel mode
		  {
			  if (sx==NUMx)
			  {
				  Pix[0][sy][sz] = Pix[NUMx][sy][sz];
				  U[NUMx+1][sy][sz] = U[1][sy][sz];
				  Ux[NUMx+1][sy][sz] = Ux[1][sy][sz];
				  Uy[NUMx+1][sy][sz] = Uy[1][sy][sz];
			  };
		  };
		  if (sy==NUMT)
		  {
			  Piy[sx][0][sz] = Piy[sx][NUMT][sz];
			  U[sx][NUMT+1][sz] = U[sx][1][sz];
			  Ux[sx][NUMT+1][sz] = Ux[sx][1][sz];
			  Uy[sx][NUMT+1][sz] = Uy[sx][1][sz];
		  };
		  if (sz==NUM)
		  {
			  Ez[sx][sy][0] = Ez[sx][sy][NUM];
			  U[sx][sy][NUM+1] = U[sx][sy][1];
			  Ux[sx][sy][NUM+1] = Ux[sx][sy][1];
			  Uy[sx][sy][NUM+1] = Uy[sx][sy][1];
		  };

	    }
};

void doIncPart2(double eps) {
  //SUNmatrix c,d;
  double step=eps/A;
  SUNmatrix temp;
  SUNLieMatrix Uc; // Ua,Ub,Uc,Ud;
  SUNLieMatrix dPlaqxz,dPlaqyz;
  SUNmatrix tempFxz,tempFyz;


    for (int sx=1;sx<=NUMx;sx++) 
	   for (int sy=1;sy<=NUMT;sy++)
	    for (int sz=1; sz<=NUM;sz++) { 
	   
	    if (COUPLE_CURRENTS) {
	    	 	
			  //cout << "doInc(): Plaq" << ",";
	
			  Uc=ux[sx][sy][sz]*u[sx+1][sy][sz]*Hermitian(ux[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
		    dPlaqxz  = Uc;
	
			  Uc=uy[sx][sy][sz]*u[sx][sy+1][sz]*Hermitian(uy[sx][sy][sz+1])*Hermitian(u[sx][sy][sz]);
		    dPlaqyz = Uc;
	
		    dPlaqxz  = (dPlaqxz + Ux[sx][sy][sz]*U[sx+1][sy][sz]*Hermitian(Ux[sx][sy][sz+1])*Hermitian(U[sx][sy][sz]))/2; 
	
		    dPlaqyz  = (dPlaqyz + Uy[sx][sy][sz]*U[sx][sy+1][sz]*Hermitian(Uy[sx][sy][sz+1])*Hermitian(U[sx][sy][sz]))/2;
	
			  tempFxz = toAlgebraElement(dPlaqxz);
			  tempFyz = toAlgebraElement(dPlaqyz);
	
		    for (int i=0;i<NW;i++) {
				  temp.reset();
				  // symmetric derivative		
		  		temp -= (transportRight(w[sx+1][sy][sz][i],ux[sx][sy][sz])-transportLeft(w[sx-1][sy][sz][i],ux[sx-1][sy][sz]))*v[i][0];
			    temp -= (transportRight(w[sx][sy+1][sz][i],uy[sx][sy][sz])-transportLeft(w[sx][sy-1][sz][i],uy[sx][sy-1][sz]))*v[i][1];
				  temp -= (transportRight(w[sx][sy][sz+1][i],u[sx][sy][sz])-transportLeft(w[sx][sy][sz-1][i],u[sx][sy][sz-1]))*v[i][2];
				  // symmetrized electric field contributions in space and time
				  temp -= (Pix[sx][sy][sz]+transportLeft(Pix[sx-1][sy][sz],ux[sx-1][sy][sz]))*a[i]*v[i][0]/2;
				  temp -= (Piy[sx][sy][sz]+transportLeft(Piy[sx][sy-1][sz],uy[sx][sy-1][sz]))*a[i]*v[i][1]/2;
				  temp += (Ez[sx][sy][sz]+transportLeft(Ez[sx][sy][sz-1],u[sx][sy][sz-1]))*(v[i][2]*a[i]+b[i])/2; 
				  temp -= (pix[sx][sy][sz]+transportLeft(pix[sx-1][sy][sz],ux[sx-1][sy][sz]))*a[i]*v[i][0]/2;
				  temp -= (piy[sx][sy][sz]+transportLeft(piy[sx][sy-1][sz],uy[sx][sy-1][sz]))*a[i]*v[i][1]/2;
				  temp += (ez[sx][sy][sz]+transportLeft(ez[sx][sy][sz-1],u[sx][sy][sz-1]))*(v[i][2]*a[i]+b[i])/2; 
		
				  temp /= 2;
				  temp += b[i]*(v[i][0]*tempFxz+v[i][1]*tempFyz);
		
				  W[sx][sy][sz][i] = ow[sx][sy][sz][i] + 2*step*temp;
				}

		  }
			else { 
				// if not coupling currents set W to zero (overkill but ok)
			  for (int i=0;i<NW;i++) W[sx][sy][sz][i].reset();
			}
			
		}

}

// => MPI implementation uses EvolveMPI()
void Evolve() {
	double eps = EPS;
	double time = 0;
	//getting started
	loadjs();
	savejs();

	outputMeasurements(time);
	for (int i=1;i<=STEPS+1;i++) {
		// evolve fields eps forward in time storing updated fields in captial vars
		doIncPart1(eps); 
		doIncPart2(eps); 
		// update hard loop energy; integrated during run
		updateHLenergy(eps); 
		// measurements and data dump
		if ( (i==2) || (i>1 && (i-1)%UPDATE==0)) {
			outputMeasurements(time);
			// Disabled for now
			//if (fabs(energyDiff)>=MAXEVIOLATION) {
			//snapshot(time);
			//break;
			//}
		}
		if ((i-1)%SNAPUPDATE==0) { snapshot(time); }
		if ((i-1)%ENERGYUPDATE==0) { dumpEnergyDensities(time); }			
		if ((i-1)%ANGLEUPDATE==0) { 
			outputSpectra(time);
			//measureangle(time); 
			}
		
		//cout << setprecision (20) << "DEBUG ENERGY - " << energyDensity() << endl;
				
		// increment time
		time = i*eps;
		// copy fields from capital vars (updated) down to lowercase vars (current)
		copyDown();
		// save currents for use in energy computation
		savejs();
	}
	


}

void printDivider() {
	int dwidth = 13;
	for (int i=0;i<10*dwidth;i++) cout << "-"; cout << endl;
	return;
}

void createDirectoryStructure()
{
	// Create directory structure
        if(mymkdir("data", 0711) == -1)
	{
		// Nothing to worry about: Most likely the directory already existed.
		// if (MSG_DISK) cout << "Could not create directory 'data_mpi'." << endl;
	}
	else
		cout << "Directory 'data' successfully created." << endl;
        
	if(mymkdir("data/snapshot", 0711) == -1)
	{
		// if (MSG_DISK) cout << "Could not create directory 'data_mpi/snapshot'." << endl;
	}	
	else
		cout << "Directory 'data/snapshot' successfully created." << endl;
};

// => MPI implementation 
//        * defines main()
//        * calls solve() if less than 1 manager + 2 workers are defined.
int solve() {

  // cout << "sizeof(SUNmatrix) = " << sizeof(SUNmatrix) 
  //		<< ", sizeof(SUNLieMatrix) = " << sizeof(SUNLieMatrix) 
  //	<< endl;

        // Set initial time for timing output
        gettimeofday(&starttime, 0);
	outputtime = starttime;
	outputcounter = 0;

	extern void readParameters(char*);
	extern void copyParameterFile(char *filename);
	
	printDivider();
	
	readParameters("params.txt");
	
	printDivider();

	allocateMemory();

	fft_init();
	
#if defined REDO_ANGLE_FROM_DUMP
	// Don't do the simulation, just read 
	// from angle dump and do angle measurement
	measureanglefromdump();
#else

	createDirectoryStructure();

	copyParameterFile("params.txt");
		
	// open files to put the data in first	
	cbar_out.open("data/cbar.dat", ios::out);
	energy_out.open("data/energy.dat", ios::out);
	gausslaw_out.open("data/gausslaw.dat", ios::out);

	initializeVelocitiesAndCoefficients();
	
	setInitialConditions();
	
  // filter fields routine from filter.cpp
  if (FILTER_FUNC>0) { 
			filterFields();
			enforceBCs();
  }

	if (PERTURB) {
  		perturbInitialConditions();
  }
    	
	printDivider();

	cout << "Simulation of " << SUNmatrix_string << endl;
		
	cout << "Spatial Step Size (A): " << A << endl;

	cout << "Temporal Step Size (EPS): " << EPS << endl;

	cout << "Standard Deviation of Initial Distribution (SIG): " << SIG << endl;

	cout << "Longitudinal Box Size (A*" << NUM << "): " << A*NUM << endl;

	cout << "Transverse Box Size (A*" << NUMT << "): " << A*NUMT << endl;

	printDivider();

	double phirms = btEnergyDensity();
	cout << "Initial Bt RMS: " << phirms << endl;

	double phizrms = bzEnergyDensity();
	cout << "Initial Bz RMS:" << phizrms << endl;

	double etrms = etEnergyDensity()*4;
	cout << "Initial et RMS: " << etrms << endl;

	double ezrms = ezEnergyDensity()*4;
	cout << "Initial ez RMS: " << ezrms << endl;

	InitialEnergy = initialED();
	cout << "Initial Energy Density: " << InitialEnergy << endl;

	cout << "Initial Lattice Energy per Plaquette: " << InitialEnergy*A*A*A*A/3 << endl;

	double gauss = checkGaussLaw();
	cout << "Initial Gauss Law Check: " << gauss << endl;

	printDivider();
	
	int dwidth = 13;
	cout.width(dwidth); cout << "Time";
	cout.width(dwidth); cout << "<Ez>";
	cout.width(dwidth); cout << "<Et>";
	cout.width(dwidth); cout << "<Bz>";
	cout.width(dwidth); cout << "<Bt>";
	cout.width(dwidth); cout << "<HL>";
	cout.width(dwidth); cout << "GaussCheck";
	cout.width(dwidth); cout << "EnergyViol";
	cout.width(dwidth); cout << "Seconds";
	cout.width(dwidth); cout << "Sec. (av)";
	cout << endl;
	
	printDivider();

	Evolve();

	printDivider();
	cout << "Done.\n";
	printDivider();

	// finally close files
	cbar_out.close();
        energy_out.close();
	gausslaw_out.close();
	
#endif //  defined REDOANGLE

	fft_end();

	return 0;
}

#ifdef STANDALONE_SOLVE

/* Solve without mpi support */
int main()
{
	solve();
}

#endif /* STANDALONE_SOLVE */
