#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "su3mat-tvmet.h"

// Static member initialization
bool SU3matrix::we_should_call_initGellMannMatrices = true;
tvmetMatrix SU3matrix::id,
	SU3matrix::s1,SU3matrix::s2,SU3matrix::s3,
	SU3matrix::s4,SU3matrix::s5,SU3matrix::s6,
	SU3matrix::s7,SU3matrix::s8; 


// Members
SU3matrix operator*(const double d,const SU3matrix &m)
{
        SU3matrix temp;
        temp.mat = m.mat;
	temp.mat *= d;
        return temp;
}

SU3matrix operator*(const int i,const SU3matrix &m)
{
        SU3matrix temp;
        temp.mat = m.mat;
	temp.mat *= i;
        return temp;
}

// Hermitian Conjugate
SU3matrix dagger(SU3matrix &m) {
	SU3matrix temp;	
	temp.mat = trans(m.mat);
	temp.mat = conj(temp.mat);
	return temp;
}

// Same function (for compatibility with SU2)
SU3matrix Hermitian(SU3matrix &m) {
	SU3matrix temp;	
	temp.mat = trans(m.mat);
	temp.mat = conj(temp.mat);
	return temp;
}


//
// transport left computes dagger(U).M.U
//
// inline SU3matrix transportLeft(SU3matrix &m,SU3matrix &u) { return dagger(u)*m*u; }

//
// transport right computes U.M.dagger(U)
//
// inline SU3matrix transportRight(SU3matrix &m,SU3matrix &u) { return u*m*dagger(u); }

double tr(SU3matrix &m,SU3matrix &n) {
	dcomp rval(0,0);
	tvmetMatrix temp; 
	temp = m.mat*n.mat; 
	rval = trace(temp);
	return real(rval);
}

double tr2(SU3matrix &m) { 
	// more efficient to code explicitly
	dcomp rval(0,0);
	tvmetMatrix temp; 
	temp = m.mat*m.mat; 
	rval = trace(temp);
	return real(rval);
}

// computes i times commuator of two SU3 matrices
SU3matrix icommute(SU3matrix &u,SU3matrix &v) {
	SU3matrix temp;
	temp.mat = (u.mat*v.mat - v.mat*u.mat)*I; 
        return temp;
}

// computes i[u,i[u,v]]
SU3matrix cuuv(SU3matrix &u,SU3matrix &v) { return icommute(u,icommute(u,v)); }

// computes matrix exponential using taylor expansion
// exp(i m) = 1 + i m + (i m)^2/2 + ...
SU3matrix mexpi(SU3matrix &m) {
        SU3matrix lm(1,0,0,0,0,0,0,0,0);
	tvmetMatrix x = m.mat;  x = x*I; 
	tvmetMatrix t; 
        t = 1,0,0,0,1,0,0,0,1; 
	tvmetMatrix n; int i;
        for (i=1;i<=TXMAX;i++) {
		t = t*x/((double)i);
		lm.mat += t;
		n = t*t;
                if (fabs(real(trace(n)))<TXACC) break;
        }
        if (i==TXMAX) cout << "WARNING:  Reached Max Terms in mexpi" << endl;
        return lm;
}

SU3matrix mexpiOld(SU3matrix &m) {
        // taylor expand exp(i m) = 1 + i m + (i m)^2/2 + ...
        SU3matrix lm(1,0,0,0,0,0,0,0,0);
        SU3matrix x;
        x.mat = m.mat*I;
        SU3matrix t;
        int i;
        for (i=1;i<=TXMAX;i++) {
                t = x;
                for (int j=2;j<=i;j++) t = x*t/j;
                lm += t;
                if (fabs(tr2(t))<TXACC) break;
        }
        if (i==TXMAX) cout << "WARNING:  Reached Max Terms in mexpi" << endl;
        return lm;
}

// convert Lie Group element to Algebra group element by subtracting the trace
SU3matrix toAlgebraElement(SU3matrix m)
{	
	tvmetMatrix id;
        id = 1,0,0,0,1,0,0,0,1;

	SU3matrix temp;
	dcomp tr(0,0);

	tr = trace(m.mat);
	temp.mat = m.mat - (id * tr / (double) 3.0);
	// Don't need factor 2 for different normalization
	// as the whole matrix is stored instead of the basis,
	// but need factor 3, as tr(id) = 3.

	return temp;
}
