/* Small test program for SU(N) */

/* Andreas Ipp, 10. July 2006 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <complex>

#include "su3mat.h"
#include "su2mat.h"

#include "random.h"

using namespace std;

typedef SU3matrix<double> SU3matrixd;
typedef SU3LieMatrix<double,complex<double> > SU3LieMatrixd;

int main()
{
  // Compare SU2 vs. SU3
  cout << " w: " << endl;
  SU2matrix w2;
  SU3matrixd w3;
  //  w2.print();
  //  w3.print();

  w2.reset();
  w2.a = .5;
  w3 = SU3matrixd(.5,0,0,0,0,0,0,0);

  w2.print();
  w3.print();

  cout << " u: " << endl;

  SU2LieMatrix u2;
  SU3LieMatrixd u3;
  //  u2.print();
  //  u3.print();

  u2.reset(); u2.a = 1;
  u3 = SU3LieMatrixd(1,0,0,0,0,0,0,0,0);

  u2.print();
  u3.print();

  // Let's do some calculation:
  cout << "mexpi" << endl;
  double step = -0.1;
  u2 = mexpi(((float)step)*w2);
  u3 = mexpi(step*w3);

  u2.print();
  u3.print();

  toAlgebraElement(u2).print();
  toAlgebraElement(u3).print();

  cout << "transportLeft" << endl;
  transportLeft(w2, u2).print();
  transportLeft(w3, u3).print();

  cout << "transportRight" << endl;
  transportRight(w2, u2).print();
  transportRight(w3, u3).print();

  cout << "Hermitian" << endl;
  Hermitian(u2).print();
  Hermitian(u3).print();
  
  cout << "tr2" << endl;
  cout << tr2(w2) << endl;
  cout << tr2(w3) << endl;
  cout << tr2(transportRight(w2,u2)) << endl;
  cout << tr2(transportRight(w3,u3)) << endl;

  cout << "tr" << endl;
  SU2matrix x2;
  SU3matrixd x3;
  x2.reset();
  x2.b = .3;
  x3 = SU3matrixd(0,.3,0,0,0,0,0,0);
  
  cout << tr(w2, x2) << endl;
  cout << tr(w3, x3) << endl;
  
  cout << "*" << endl;
  SU2LieMatrix v2;
  SU3LieMatrixd v3;

  for (int i=1; i<10; i++)
  {
	  cout << " --- " << i << endl;
    u2.reset(); u2.a = random(); u2.b = random();
    u3 = SU3LieMatrixd(u2.a,u2.b,0,0,0,0,0,0,0);
    v2.reset(); v2.a = random(); v2.b = random();
    v3 = SU3LieMatrixd(v2.a,v2.b,0,0,0,0,0,0,0);
    //v3.m12 = I * 1.; v3.m21 = I * 1.;
   u2.print();
   v2.print();
    
  (u2 * v2).print();
  (u3 * v3).print();
  (v2 * u2).print();
  (v3 * u3).print();
  toAlgebraElement(u2 * v2).print();
  toAlgebraElement(u3 * v3).print();
  };   

#if defined WIN32
	// Under Windows, we wait for any input to prevent the window
	// to close before we can study the output:
	char dummy[255];
//	cin >> dummy;
#endif
	exit(0);
};
