/* Small test program for SU(N) */

/* Andreas Ipp, 10. July 2006 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <complex>

#include "sunmat.h"

//#include "su3mat.h"
#include "random.h"

using namespace std;

void test_su3()
{
#if defined SU3
	cout << "Starting test..." << endl;

	SU3matrix<double> m1(1,2,3,4,5,6,7,8);
	SU3matrix<double> m2, m3;

	m2.reset();
	m2.set(11,12,13,14,15,16,17,18);

	m3 = m1 + 2 * m2;

	m1.print();
	m2.print();
	m3.print();

	complex<double> a;

	SU3LieMatrix< double, complex<double> > L1, L2, L3;
	
	L1 = m1;
	L1.print();
	L2 = m2; 
	L3 = m3;

	m1 = toAlgebraElement(L1);
	m1.print();

	m1 = toAlgebraElement(L2);
	m1.print();

	m1 = toAlgebraElement(L3);
	m1.print();

	L3 = L1 * L2;
	cout << "L3 = L1 * L2:" << endl;
	L3.print();
	toAlgebraElement(L3).print();

	L3 = L2 * L1;
	cout << "L3 = L2 * L1:" << endl;
	L3.print();
	toAlgebraElement(L3).print();

	cout << "mexpi(0.1*m1):" << endl;
	m2 = 0.1 * m1;
	L3 = mexpi(m2);
	L3.print();

	cout << "Commute: " << endl;
	m1.set(1,2,3,4,5,6,7,8);
	m2.set(11,12,13,14,15,16,17,18);

	icommute(m1,m2).print();
	icommuteSLOW(m1,m2).print();
#endif
}

void test_sun_old01()
{
	cout << "Starting test..." << endl;

	SUNmatrix m1(1,2,3,4,5,6,7,8);
	SUNmatrix m2, m3;

	m2.reset();
	m2 = SUNmatrix(11,12,13,14,15,16,17,18);

	m3 = m1 + 2 * m2;

	m1.print();
	m2.print();
	m3.print();

	complex<double> a;

	SUNLieMatrix L1, L2, L3;
	
	L1 = m1;
	L1.print();
	L2 = m2; 
	L3 = m3;

	m1 = toAlgebraElement(L1);
	m1.print();

	m1 = toAlgebraElement(L2);
	m1.print();

	m1 = toAlgebraElement(L3);
	m1.print();

	L3 = L1 * L2;
	cout << "L3 = L1 * L2:" << endl;
	L3.print();
	toAlgebraElement(L3).print();

	L3 = L2 * L1;
	cout << "L3 = L2 * L1:" << endl;
	L3.print();
	toAlgebraElement(L3).print();

	cout << "mexpi(0.1*m1):" << endl;
	m2 = 0.1 * m1;
	L3 = mexpi(m2);
	L3.print();

	cout << "Commute: " << endl;
	m1 = SUNmatrix(1,2,3,4,5,6,7,8);
	m2 = SUNmatrix(11,12,13,14,15,16,17,18);

	icommute(m1,m2).print();
#if defined SU3
	icommuteSLOW(m1,m2).print();
#endif
}

void test_sun()
{
	SUNmatrix m1, m2, m3;
	SUNLieMatrix L1, L2, L3;

	cout << endl << "----------" << endl;
	cout << SUNmatrix_string << endl;
	cout << "Test Hermitian" << endl;
	m1 = SUNmatrix(1,2,3,4,5,6,7,8);
	m2 = SUNmatrix(11,12,13,14,15,16,17,18);
	L1 = SUNLieMatrix(1,0,0,0,0,0,0,0,0);
	//L1.print();

	L1 = m1;
	L2 = Hermitian(L1);
	L3 = L1 * L2;

	//L1.print();
	//L2.print();
	//L3.print();

	m1 = toAlgebraElement(2 * L2 * L3);
	L1 = m1;
	//L1.print();

	cout << " ----> transportRight <----" << endl;
	m1 = transportRight(m2, L2)/2;
	L1 = m1;
	//L1.print();

	L1 = dagger(L2);
	//L1.print();

	L3 = m2;
	L1 = L3*dagger(L2);
	//L1.print();

	L1 = L2 * L1;
	//L1.print();

	L1 = L1 / 2;
	//L1.print();

	L1 = L2 * L3;
	//L1.print();

	L1 = L1 * dagger(L2) / 2;
	//L1.print();

	L1 = L2 * L3 * dagger(L2) / 2;
	//L1.print();

	m1 = toAlgebraElement(L1);
	L1 = m1;
	//L1.print();

	cout << " ----> icommute <----" << endl;
	m1 = SUNmatrix(1,2,3,4,5,6,7,8);
	m2 = SUNmatrix(11,12,13,14,15,16,17,18);

	m3 = icommute(m1, m2);

	L3 = m3;
	L3.print();

#if defined SU3 || defined SU3double
	cout << "  SLOW:" << endl;
	m3 = icommuteSLOW(m1, m2);

	L3 = m3;
	L3.print();
#endif

}

int main()
{
#if defined SU3
	test_su3();
#endif

	test_sun();


#if defined WIN32
	// Under Windows, we wait for any input to prevent the window
	// to close before we can study the output:
	char dummy[255];
//	cin >> dummy;
#endif
	exit(0);
};
